
// *****************************************************************************
//
// source:    utdevt.C
//
// created:   11 May 1999
//
// Author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// symbol:    utdevt_()
//
// called by: mmrupd.F
//
// purpose:   build DNNN tables from DEVT
//
// changes:
//                  final polishing (AG: 12 May 99)
//                  move over to ArteTable::size() instead of nrow() (AG: 31 May 99)
//                  put includes (AG: 1 Jun 99)
// 09 Sep 1999 (AG) replace cerr by cout
// 09 Nov 1999 (AG) removed a warning
//
// *****************************************************************************

// includes
#include <iostream>

#include "grr.h"
#include "event.h"

#include "arte/arte.hh"

#include "farm/fa.hh"
#include "farm/farm.hh"

// C prototpyes
extern "C" void mmbook_( int & , const int & );
extern "C" void mmfree_( int & );

// externals
extern int FADEB;

//
// program code
//
extern "C" void
utdevt_()
{
  // DEVT
  EVENT_HDR *event = ARTE_get_event(); // event data pointer
  if( !event ) return;                 // check pointer

  // event header
  EVENT_HDR evtHdr;
  ESP_get_event_hdr( event, &evtHdr );
#ifdef _DEB_
  if( FADEB ){
    cout << endl;
    cout << "utdevt_(): READ EVENT: ********************************************"
	 << endl;
    cout << "length="            << evtHdr.length          << " "
	 << "recid=0x"    << hex << evtHdr.recid    << dec << " "
	 << "format=0x"   << hex << evtHdr.format   << dec << " "
	 << "sltevt="            << evtHdr.sltevt          << " "
	 << "compmask=0x" << hex << evtHdr.compmask << dec << " "
	 << "evtmask=0x"  << hex << evtHdr.evtmask  << dec << endl;
    cout << "*******************************************************************"
	 << endl;
    cout << endl;
  }
#endif
 
  // DSVD (needs collection of FED blocks first)
  int  nDSVD   = 50000;
  int  lenDSVD = 0;
  char *pDSVD  = (char *) malloc( nDSVD );
  
  // inits
  int        rc    = GRR_OK;
  static int count = 0;
  count++;
  const int max = 1000; // maximal SLB records
  SLB_HDR   *slbp[max]; // array of SLB records

  // get pointers to SLBs of COMP_ID
  int n = 0;
  if( ( rc = ETB_get_slb_points( event, -1, max, slbp, &n ) ) == ETB_ok ){
#ifdef _DEB_
    if( FADEB ){
      cout << "utdevt(): ETB_get_slb_points() n=" << n << endl;
    }
#endif
  }
  else {
#ifdef _ERR_
    cout << "utdevt(): ERROR: ETB_get_slb_points() rc=" << rc << endl;
#endif
    return;
  }
  
  // SLB pointer loop
  for( int i=0; i<n; i++ ){
    SLB_HDR slbHdr;
    ESP_get_slb_hdr( (SLB_HDR *) slbp[i], &slbHdr ); // translate SLB_HDR

    // table
    int  recid  = slbHdr.recid;
    int  compid = slbHdr.compid;
    int  slbid  = slbHdr.slbid;
    int  iName  = faID2Name( compid );
    int  size   = slbHdr.length - sizeof( SLB_HDR ) - sizeof( SLB_TAIL );
    char *pData = (char *) slbp[i] + sizeof( SLB_HDR  );

    // print
#ifdef _DEB_
    if( FADEB ){
      cout << endl;
      cout << "utdevt(): SLB: i=" << i                     << " "
	   << "length="          << slbHdr.length          << " "
	   << "recid=0x"  << hex << slbHdr.recid    << dec << " "
	   << "fltevt=0x" << hex << slbHdr.fltevt   << dec << " "
	   << "compid=0x" << hex << slbHdr.compid   << dec << " "
	   << "slbid=0x"  << hex << slbHdr.slbid    << dec << " "
	   << "nodeid=0x" << hex << slbHdr.nodeid   << dec << " "
	   << "name="            << faName( iName )        << " "
	   << "size="            << size                   << endl;
    }
#endif
    
    // check
    if( count < 11 ){
      if( size == 0 ){
#ifdef _OUT_
	cout << "  "
	     << count << " "
	     << "WARNING (this message will stop after 10 events): "
	     << "SLB record has NO data: "
	     << "compid=0x" << hex << compid << dec << " "
	     << "slbid=0x"  << hex << slbid  << dec << endl;
#endif
      }
    }

    //
    // handle SLB recods
    //
    int nSparse = 0;

    //
    // DDAQ 0x0000 *************************************************************
    //
    if( iName == faName2Int( "DDAQ" ) ){
      int nrow = ArteTable<DDAQ>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DDAQ: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DDAQ>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DDAQ>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DDAQ> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DDAQ: "
	     << "newID=0x"  << hex << *newID << dec << " "
	     << "nrow="     << ArteTable<DDAQ>::size() << " "
	     << "nSparse="  << nSparse         << " "
	     << "red="      << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DDAQ>::getLink().getRef(), red );
	mmfree_( ArtePointer<DDAQ>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DDAQ: nrow()=" << ArteTable<DDAQ>::size() << endl;
      }
#endif
    }

    //
    // DSVD 0x0100 (first collect all FED blocks) ******************************
    //
    if( iName == faName2Int( "DSVD" ) ){
      if( ( lenDSVD + size ) > nDSVD ){
	pDSVD = (char *) realloc( pDSVD, lenDSVD + size );
	nDSVD = lenDSVD + size;
      }
      memcpy( pDSVD+lenDSVD, pData, size );
      lenDSVD += size;
    }

    //
    // DITR 0x0200 *************************************************************
    //
    if( iName == faName2Int( "DITR" ) ){
      int nrow = ArteTable<DITR>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DITR: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DITR>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DITR>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DITR> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DITR: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << ArteTable<DITR>::size() << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DITR>::getLink().getRef(), red );
	mmfree_( ArtePointer<DITR>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DITR: nrow()=" << ArteTable<DITR>::size() << endl;
      }
#endif
    }
    
    //
    // DPTC 0x0300 *************************************************************
    //
    if( iName == faName2Int( "DPTC" ) ){
      int nrow = ArteTable<DPTC>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DPTC: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DPTC>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DPTC>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DPTC> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DPTC: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << ArteTable<DPTC>::size() << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DPTC>::getLink().getRef(), red );
	mmfree_( ArtePointer<DPTC>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DPTC: nrow()=" << ArteTable<DPTC>::size() << endl;
      }
#endif
    }

    //
    // DOTR 0x0400 *************************************************************
    //
    if( iName == faName2Int( "DOTR" ) ){
      int nrow = ArteTable<DOTR>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DOTR: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DOTR>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DOTR>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DOTR> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DOTR: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << ArteTable<DOTR>::size() << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DOTR>::getLink().getRef(), red );
	mmfree_( ArtePointer<DOTR>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DOTR: nrow()=" << ArteTable<DOTR>::size() << endl;
      }
#endif
    }

    //
    // DRIC 0x0500 *************************************************************
    //
    if( iName == faName2Int( "DRIC" ) ){
      int nrow = ArteTable<DRIC>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DRIC: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DRIC>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DRIC>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DRIC> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DRIC: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << ArteTable<DRIC>::size() << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DRIC>::getLink().getRef(), red );
	mmfree_( ArtePointer<DRIC>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DRIC: nrow()=" << ArteTable<DRIC>::size() << endl;
      }
#endif
    }

    //
    // DTRD 0x0600 *************************************************************
    //
    if( iName == faName2Int( "DTRD" ) ){
      int nrow = ArteTable<DTRD>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTRD: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DTRD>::getLink().getRef(),
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DTRD>::getLink().getRef(),
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DTRD> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTRD: "
	     << "newID=0x"  << hex << *newID   << dec << " "
	     << "nrow()="          << ArteTable<DTRD>::size() << " "
	     << "nSparse="         << nSparse         << " "
	     << "red="             << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DTRD>::getLink().getRef(), red );
	mmfree_( ArtePointer<DTRD>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTRD: nrow()=" << ArteTable<DTRD>::size() << endl;
      }
#endif
    }

    //
    // DCAL 0x0700 *************************************************************
    //
    if( iName == faName2Int( "DCAL" ) ){
      int nrow = ArteTable<DCAL>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DCAL: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DCAL>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DCAL>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DCAL> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DCAL: "
	     << "newID=0x" << hex << *newID   << dec << " "
	     << "nrow()="         << ArteTable<DCAL>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DCAL>::getLink().getRef(), red );
	mmfree_( ArtePointer<DCAL>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DCAL: nrow()=" << ArteTable<DCAL>::size() << endl;
      }
#endif
    }

    //
    // DMUT 0x0800 *************************************************************
    //
    if( iName == faName2Int( "DMUT" ) ){
      int nrow = ArteTable<DMUT>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DMUT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DMUT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DMUT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DMUT>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DMUT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DMUT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUT: nrow()=" << ArteTable<DMUT>::size() << endl;
      }
#endif
    }

    //
    // DMUX 0x0900 *************************************************************
    //
    if( iName == faName2Int( "DMUX" ) ){
      int nrow = ArteTable<DMUX>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUX: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DMUX>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DMUX>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DMUX> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUX: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DMUX>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DMUX>::getLink().getRef(), red );
	mmfree_( ArtePointer<DMUX>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUX: nrow()=" << ArteTable<DMUX>::size() << endl;
      }
#endif
    }

    //
    // DMUP 0x0a000 *************************************************************
    //
    if( iName == faName2Int( "DMUP" ) ){
      int nrow = ArteTable<DMUP>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUP: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DMUP>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DMUP>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DMUP> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUP: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DMUP>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DMUP>::getLink().getRef(), red );
	mmfree_( ArtePointer<DMUP>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DMUP: nrow()=" << ArteTable<DMUP>::size() << endl;
      }
#endif
    }

    //
    // DFLT 0x0b00 *************************************************************
    //
    if( iName == faName2Int( "DFLT" ) ){
      int nrow = ArteTable<DFLT>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DFLT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DFLT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DFLT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DFLT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DFLT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DFLT>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DFLT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DFLT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DFLT: nrow()=" << ArteTable<DFLT>::size() << endl;
      }
#endif
    }

    //
    // DSLT 0x0c00 *************************************************************
    //
    if( iName == faName2Int( "DSLT" ) ){
      int nrow = ArteTable<DSLT>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DSLT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DSLT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DSLT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DSLT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DSLT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DSLT>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DSLT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DSLT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DSLT: nrow()=" << ArteTable<DSLT>::size() << endl;
      }
#endif
    }

    //
    // DTLT 0x0d00 *************************************************************
    //
    if( iName == faName2Int( "DTLT" ) ){
      int nrow = ArteTable<DTLT>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTLT: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DTLT>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DTLT>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DTLT> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTLT: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DTLT>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DTLT>::getLink().getRef(), red );
	mmfree_( ArtePointer<DTLT>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTLT: nrow()=" << ArteTable<DTLT>::size() << endl;
      }
#endif
    }

    //
    // DTAR 0x0e00 *************************************************************
    //
    if( iName == faName2Int( "DTAR" ) ){
      int nrow = ArteTable<DTAR>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTAR: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DTAR>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DTAR>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DTAR> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTAR: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DTAR>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DTAR>::getLink().getRef(), red );
	mmfree_( ArtePointer<DTAR>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DTAR: nrow()=" << ArteTable<DTAR>::size() << endl;
      }
#endif
    }

    //
    // DFCS 0x0f00 *************************************************************
    //
    if( iName == faName2Int( "DFCS" ) ){
      int nrow = ArteTable<DFCS>::size();
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DFCS: "
	     << "nrow=" << nrow << " "
	     << "size=" << size << endl;
      }
#endif
      // book
      if( !nrow ){
	mmbook_( ArtePointer<DFCS>::getLink().getRef(), 
		 size / sizeof( int ) + 1 );
	nrow = 1; // jump to second word
      }
      else {
	mmbook_( ArtePointer<DFCS>::getLink().getRef(), 
		 size / sizeof( int ) );
      }
      // table
      ArtePointer<DFCS> t( 1 );
      int  *newID = (int *)   &t->bits;    
      char *p     = (char *)( &t->bits + nrow );
      // sparsify
      nSparse = faCopySlb( &recid, pData, newID, p, size );
      int red = ( nSparse - size ) / int( sizeof( int ) );
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DFCS: "
	     << "newID=0x" << hex << *newID   << dec << " "
             << "nrow()="         << ArteTable<DFCS>::size() << " "
	     << "nSparse="        << nSparse         << " "
	     << "red="            << red             << endl;
      }
#endif
      if( red ){
	mmbook_( ArtePointer<DFCS>::getLink().getRef(), red );
	mmfree_( ArtePointer<DFCS>::getLink().getRef() );
      }
#ifdef _DEB_
      if( FADEB ){
	cout << "utdevt(): DFCS: nrow()=" << ArteTable<DFCS>::size() << endl;
      }
#endif
    }
  }

  //
  // build DSVD
  //
  if( lenDSVD ){
    int recID = 0x2000;
    mmbook_( ArtePointer<DSVD>::getLink().getRef(),
	     lenDSVD / sizeof( int ) + 1 );
    ArtePointer<DSVD> t( 1 );
    int  *newID = (int *)   &t->bits;    
    char *p     = (char *)( &t->bits + 1 );
    // sparsify?
    int nSparse = faCopySlb( &recID, pDSVD, newID, p, lenDSVD );
    int red     = ( nSparse - lenDSVD ) / int( sizeof( int ) );
#ifdef _DEB_
    if( FADEB ){
      cout << endl;
      cout << "utdevt(): DSVD: "
	   << "newID=0x" << hex << *newID   << dec << " "
	   << "nrow()="         << ArteTable<DSVD>::size() << " "
	   << "nSparse="        << nSparse         << " "
	   << "red="            << red             << endl;
    }
#endif
    if( red ){
      mmbook_( ArtePointer<DSVD>::getLink().getRef(), red );
      mmfree_( ArtePointer<DSVD>::getLink().getRef() );
    }
  }

  // free memory
  free( pDSVD );
}
