//                              -*- Mode: C++ -*-
//:ECBuilder.C
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 11/4/97
// Last Update ........
//
//

#include "care/ECBuilder.hh"
#include "care/ELNearClusters.hh"
#include "care/ELNearHitCells.hh"
#include "arte/ArteMessage.hh"
#include "care/EMClusterTree.hh"
#include <algorithm>

#undef WRITE

/*
CLASS
  ECBuilder
  Provide a collection of basic function for the building process
  of the hierarchical cluster tree. 

*/



////////////////// Class ECBuilder


//--
// Prepare internal data for a loop on all cell. It does the 
// sorting of EMHitCellArray.
void ECBuilder::startClustering(ECHitCellArray & signalCells, 
				EMClusterTree & topClusterList){
  pHitCells = &signalCells; 
  pTopClusterList = &topClusterList;
  // Sorts in energy the array of pointers
  pHitCells->sortCells();
  // set iterator to first cell
  currHitCell = pHitCells->begin();
  // inform the Cluster Tree that a clusterization process has started
  pTopClusterList->start(pHitCells->size());
}


//--
// Do clusterize with current cell.
// Insert it in an old cluster or in a new one.  
void ECBuilder::doClusterize(){

  // consider the ID of the current Hit Cell
  //   ECCellID currID = ArtePointer<HITC>(*currHitCell)->gcce;
  // transform it to the current cell index
  currIdx = ArtePointer<HITC>(*currHitCell)->gcce;
  //
  // start a EMHitCell type and store to thw whichCluster LUT
  // set the pointer to the current cell
  EMHitCell* hitcell = pTopClusterList->newHitCell(currIdx);
  hitcell->pointToCell(ArtePointer<HITC>(*currHitCell));
# ifdef WRITE
  std::cout<<" Clustering cell Idx:"<<currIdx.index()<<std::endl;
# endif

  // Do clusterization:
  // Get the number of nearest neighboring clusters
  //  buildNearestClusterArray();  
  nearClus.of(currIdx, *pTopClusterList);
# ifdef WRITE
  std::cout<<" Found "<<nearClus.howMany()<< " nearby clusters "<<std::endl;
# endif
  //  if( numberNNC==0 ){
  if( nearClus.howMany()==0){
    // the array is empty: start a new top cluster with this cell
#   ifdef WRITE
    std::cout<<" Adding to the Top Cluster List"<<std::endl;
#   endif
    pTopClusterList->newCluster(hitcell)->appendTo(pTopClusterList);
#   ifdef WRITE
    std::cout<<" added! "<<std::endl;
#   endif
    //
  } else {
    if( nearClus.howMany()>1 ){
      // there is more than one cluster. Find minimum set compatible
      nearClus.minimize(); 
#     ifdef WRITE
      std::cout<<" Minimized list contains "<<nearClus.howMany() <<" clusters."<<std::endl;
#     endif
    }
    if(nearClus.howMany()==1){
      // only a cluster available: add this cell to this cluster
      nearClus.begin();
      hitcell->appendTo(*nearClus);
    } else {
#     ifdef WRITE
      std::cout<<" Building a complex cluster: starting situation "<<std::endl;
	pTopClusterList->dump();
#     endif
      // more than one cluster: they share the same parent 
      // (i.e. they belong to the TopCluster list).
      // Merging require ordering in energy: assume it is given by minimize 
      // for NNCM and by the following function after
      // a new cluster can be found using the near Clusters and this one
      nearClus.sortClusters();
      // Now start a main cluster.
      // The new cluster will always be added to the parent cluster
      nearClus.begin();
#     ifdef WRITE
      std::cout<<" Creating a new cluster with depth "
	  <<(*nearClus)->depthLevel()<<std::endl;
#     endif

      EMCluster* pClus = pTopClusterList->newCluster(*nearClus);
      nearClus.next();
      // move all neibouring clusters
      for(; nearClus.more(); nearClus.next()){  
	//(*nearClus)->appendTo(pClus);
	pClus->moveAtEnd(*nearClus); 
      }
      // add the current cell
      hitcell->appendTo(pClus);
#     ifdef WRITE
      std::cout<<" Final situation "<<std::endl;
      pTopClusterList->dump();
#     endif
    }
  }
# ifdef WRITE
  pTopClusterList->dump();
# endif
}

//--
/// start a new cluster with a given cell. The passed cell is actually MOVED
EMCluster * ECBuilder::newCluster(EMClusterMember * clusmem) const{
  return pTopClusterList->newCluster(clusmem);
} 



//--
// Perform the Track-Cluster Match
void ECBuilder::trackClusterMatch(){
  EMHitCell* where;
  EMCluster*  clust;

  for(std::vector<EMTrackOnEcal>::iterator track
		=pTopClusterList->beginTrackList();
      track!=pTopClusterList->endTrackList(); track++){
    // extrapolate to the first plane of ECAL
    where = track->extrapolate(pTopClusterList);
    if(where!=0){
      // get the parent of the cell
      clust = where->getParent(); 
      // add the track to current cluster
      track->appendTo(clust);
    }
  }
}





//--
// Try to recover unmatched tracks
void ECBuilder::recoverUnmatchedTracks(){
  // try to recover tracks that are not connected to any cluster
  // start selecting unmatched tracks
  // find the ECAL cell in which the track is (if any)
  //
  CellIdx_t cell;

  EMHitCell* where;

  for(std::vector<EMTrackOnEcal>::iterator track
		=pTopClusterList->beginTrackList();
      track!=pTopClusterList->endTrackList(); track++){
    where = track->whichCell();
    if(where==0){
      // 
      // here trackOnEcal[i] is an unmatched track
      //
      cell = track->whichCellIdx();
      //
      if( cell.index()>0){
        //
	// the track points somewhere on ECAL
	// 
	// Find if any of the neibouring cells have been clusterized.
	//
	EMCluster* clusmem = findACluster(track, cell);
	//
	// a base cluster (or a single cell) has been found
	if(clusmem!=0){
	  // append the track
	  track->appendTo(clusmem);
     	  //
	  // update cluster hypotheses
	  //
	  int neutr = clusmem->getNeuHyp();
	  EMCluster* parent = clusmem;
	  while( parent!=0){
	    // for each parent cluster
	    // there is a true track in addition
	    // parent->newTrack();
	    // there is a track hypothesis
	    parent->setChargHyp(parent->getChargHyp()+1);
	    // there is 'neutr' neutrals less (neutr = 0 or 1)
	    if( neutr!=0)
	      parent->setNeuHyp(parent->getNeuHyp()-neutr);
	    else
	      // there is a general hypotheses more
	      parent-> setHypotheses(parent->getHypotheses()+1);
	    parent = parent->getParent();
	  }
	}
      }
    }
  }
}


//--
/// fill a fraction of cell and store it in the tree
void ECBuilder::cellFraction(EMHitCell * cell, EMCluster* clust, 
			     Energy_t fraction){
  // ask for a new cell fraction
  EMCellFraction* where = pTopClusterList->newCellFraction(cell, fraction);
  //std::cout<<"Adding a fractional cell:"<<std::endl;
  //where->dump();
  //std::cout<<" to the cluster: ("<<clust<<")"<<std::endl;
  //clust->dump();
  //std::cout<<" Appending "<<std::endl;
  // store in the cluster "cluster"
  if( where!=0 )
    where->appendTo(clust);
}




//--
//  return a pointer to a cluster in which the current cell could be
//  inserted in substitution to the cluster found by extrapolate.
EMCluster*  ECBuilder::findACluster(EMTrackOnEcal* track, 
				    CellIdx_t cell) const {
  //
  // cut off distances
  //
  static float cutOffDistancesq1 = 560.; // max diag 1+0.5 mod sq
  //   
  EMCluster*  clus = 0;

  if( cell.index()>0){ // consider only Track pointing to some cell
    //
    // start taking the list of nearly clusterized cells
    //
    ELNearHitCells nearCells;
    nearCells.of(cell,*pTopClusterList);
    EMHitCell* currCell;
    //
    float distsq = 1000000.; // this is a 1000 cm distance
    EMHitCell* NearestCell = 0;
    // Now loop on all hitCellArr and select only the
    // base clusters having at least one cell in the array
    // and are classified neutral or void.
    // Then evaluate distance between the extrapolated point on the
    // calorimeter and the peak of the cluster
    for(nearCells.begin(); nearCells.more(); nearCells.next()){
      currCell = *nearCells;
      EMCluster* parent =  currCell->getParent();
      //
      // skip clusters that are not base one
      //
      if( parent->begin()->isALeaf() && parent->getTracks()==0 &&
	  parent->getNeuHyp()==0){
	// I do not like this cast !!!!!!!!!!!!!!!!!!!
	EMHitCell * child = (EMHitCell*) parent->begin();
        // now select only void clusters or single cells
        // 
	//
	// here currCell can be a single cell or a base cluster
	//
	float dist = track->distancesq(child);
	if( dist<distsq ){
	  NearestCell = child;
	  distsq = dist;
	}
      }
    }
#ifdef HIST
    if(NearestCell!=0)
      h1d[ctd3]->accumulate(distsq);
#endif
    if( distsq<cutOffDistancesq1 ){
      // cluster with track 
      clus = NearestCell->getParent();
    } else {
      clus = 0;
    }
  }
  return clus;
}



//--
//  return a pointer to a cluster in which the current cell could be
//  inserted in substitution to the cluster found by extrapolate.
EMCluster*  ECBuilder::findOtherCluster(EMTrackOnEcal* track) const {
  // cut off distances
  //
  static float cutOffDistancesq1 = 560.; // max diag 1+0.5 mod sq
  static float cutOffDistancesq2 = 1554.; // max diag 2+0.5 mod sq
  //
  EMCluster* nextCluster = 0;
  if( track->whichCell()!=0){ 
    // consider only Track pointing to some clustered Hit
    EMCluster* cluTrack = track->getParent();
    //
    ELNearHitCells nearCells; 
    nearCells.of(track->whichCell(), *pTopClusterList);
    ELNearClusters nearClusters(nearCells);
    //
    EMHitCell* currCell;
    EMCluster*  clusList[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    unsigned int  clusnum = 0;
    int  clusnear =-1;
    float distsq = 1000000.; // this is a 1000 cm distance
    // Now loop on all hitCellArr and select only the
    // base clusters having at least one cell in the array
    // and are classified neutral or void.
    // Then evaluate distance between the extrapolated point on the
    // calorimeter and the peak of the cluster
    //
    for(nearClusters.begin(); nearClusters.more(); nearClusters.next()){
      EMCluster* parent =  *nearClusters;
      // skip clusters not below the current tree
      if( cluTrack==(parent->upstream(cluTrack->depthLevel()))){
	EMClusterMember* child = parent->begin();
	if( child->isALeaf() ){
	  // the cell is in a base Cluster;
	  // look if it is a neutral or a void cluster
	  if( parent->getTracks()==0 && parent->getNeuHyp()<2){
	    // the cluster is a neutral or void: add it to the
	    // cluster list if it is not already present
	    unsigned int i;
	    for(i=0; i<clusnum; i++)
	      if( clusList[i] == parent) break;
	    if( i==clusnum){
	      // the current cluster is NOT present in the cluster list: add it
	      clusList[clusnum] = parent;
	      float actualdistsq = track->distancesq((EMHitCell*)child);
	      if( actualdistsq<distsq ){
		// this cluster is nearer of the previous found (if any)
		// store it!
		distsq = actualdistsq;
		clusnear = clusnum;
	      }
	      clusnum++;
	    }
	  }	
	}
      }
    }
#ifdef HIST
    if(clusnear>=0)
      h1d[ctd1]->accumulate(distsq);
#endif
    if( clusnear>=0 && distsq < cutOffDistancesq1 ){ 
      // a near void or neutral cluster has been found
      nextCluster = clusList[clusnear];
    } else {
      // the cell is not directly near any void or neutral cluster
      // it is still possible that the cell is in the neibourhood of
      // a neutral or void cluster. I have to analyze all the cluster
      // structure in which neibouring cells are located.
      std::vector<EMCluster*> baseClust; 
      std::vector<EMHitCell*> bouRegCell;
      unsigned int i;
      for(i=0; i<clusnum; i++)
	baseClust.push_back(clusList[i]); // remember clusters already 
      i= 0;
      for(nearCells.begin(); nearCells.more(); nearCells.next()){
	currCell = *nearCells;
	EMCluster* parent =  currCell->getParent();
	EMClusterMember* child = parent->begin();
	if( !child->isALeaf() )
	  bouRegCell.push_back(currCell); // remember cells in bound.Reg.
      }
      // 
      // for each cell in the boundary region:
      // find the near cells
      for(i=0; i<bouRegCell.size(); i++){
	ELNearHitCells nearNearCells;
	nearNearCells.of(bouRegCell[i], *pTopClusterList);
	// for each near cell:
	for(nearNearCells.begin(); nearNearCells.more(); 
	    nearNearCells.next()){
	  currCell = *nearNearCells;
	  EMCluster* parent =  currCell->getParent();
	  // skip clusters not below the current tree
	  if( cluTrack==currCell->upstream(cluTrack->depthLevel())){
	    EMClusterMember* child = parent->begin();
	    if( !child->isALeaf() ){
	      // test if it is present in the list
	      std::vector<EMHitCell*>::const_iterator iter = bouRegCell.begin(); 
	      while (iter != bouRegCell.end() && *iter != currCell) ++iter;
	      if( iter== bouRegCell.end() ){
		bouRegCell.push_back(currCell); // remember cells in 
	      // bound.Reg.
	      }
	    } else {
	      if( parent->getTracks()==0 && parent->getNeuHyp()<2){
		// the cluster is a neutral or void: add it to the
		// cluster list if it is not already present
		std::vector<EMCluster*>::const_iterator iter(baseClust.begin());
		while (iter != baseClust.end() && *iter != parent) ++iter;
		if( iter== baseClust.end() ){
		  baseClust.push_back(parent); // remember new parent cluster 
		  clusnum++;
		}
		// analyze parent cluster
		float actualdistsq = track->distancesq((EMHitCell*)child);
		if( actualdistsq<distsq ){
		  // this cluster is nearer of the cluster previous found
		  // store it
		  distsq = actualdistsq;
		  clusnear = clusnum-1;
		}
	      }
	    }
	  }
	}
      }
      // here distancesq is the minimum distance found from all the 
      // near base clusters
#ifdef HIST
      if( clusnear>=0 )
	h1d[ctd2]->accumulate(distsq);
#endif
      if( clusnear>=0 && distsq<cutOffDistancesq2 ){
	 nextCluster = baseClust[clusnear];
      } else {
	nextCluster = 0;
      }
    }
  } 
  return nextCluster;
}









