
// *****************************************************************************
//
// source:    faInit.C
//
// type:      source code
//
// created:    7 Jul 1997
//
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// symbol:    fainit_()
//
// input:     void
// return:    int  ARTE_error_code
//
// called by: ATINIT
//
// purporse:  initialization of farm stuff called for FARM mode only
//
// changes:         
//                  initialize and register rpm (AG: 25-Nov-98)
//                  multiple file input (AG: 27-Nov-98)
//                  daq format logging (AG: 1-Dec-98)
//                  includes (AG: 11 Jan 99)
//                  removed emg status flag (AG: 12 Jan 99)
//                  obey ANSI rules (AG: 15 Jan 99)
//                  call logger initialization at the end (AG: 21 Jan 99)
//                  rpm register (AG: 2 Feb 99)
//                  some printing removed (AG: 10 Feb 99)
//                  removed faUtil.hh (AG: 23 Feb 99)
//                  use grr for error logging (AG: 25 Feb 99)
//                  polishing (AG: 11 Mar 99)
//                  put includes (AG: 1 Jun 99)
// 18 Aug 1999 (AG) polishing
// 09 Sep 1999 (AG) always use cout
// 08 Oct 1999 (AG) write start-up to log-file
// 18 Oct 1999 (AG) add command line arguments to log-file output
// 01 Nov 1999 (AG) re-try rpm_open() if it fails
// 03 Nov 1999 (AG) add process id to logfile
// 11 Nov 1999 (AG) improve log file
// 24 Nov 1999 (AG) use ARTE_error()
// 24 Nov 1999 (AG) write settings into logfile
// 25 Nov 1999 (AG) introduce SMC initialization
// 29 Nov 1999 (AG) print the configuration into the logfile
// 01 Dec 1999 (AG) finish logfile output
// 06 Dec 1999 (AG) initialize pinging
// 07 Dec 1999 (AG) publish event count
// 10 Dec 1999 (AG) publish more quantities
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

// ONLINE includes
#include "rpm.h"
#include "rpi.h"
#include "smc.h"
#include "rhp.h"
#include "grr.h"
#include "fld.h"
#include "par.h"
#include "log.h"
#include "runcfg.h"

// ARTE includes
#include "arte/ArteFlag.h"
#include "farm/fa.hh"
#include "farm/farm.hh"

// globals
char *evtBuf   = 0;  // event buffer
int  nodeID    = -1; // default 4LT node ID

// externals
extern int  nArg;   // line command arguments
extern char **vArg; // line command arguments

extern int  FADEB;    // debug flag
extern int  FARM;     // FARM mode
extern int  FAON;     // FARM online mode
extern char *DBNAME;  // database name
extern char *LOGFILE; // log file
extern int  NOSPAR;   // sparsification
extern int  RPMINIT;  // init rpm/rhp
extern int  DAQLOG;   // daq format logging
extern int  FAGRR;    // error logger
extern char *PARNAME; // emg partition
extern int  SEG_NUM;  // emg partitions
extern int  SEG_LEN;  // emg partition size

// C++ prototypes
void faCatchSIG( int sig );
SMC_CODES faSMCinit2conf ( void *dummy1, void *dummy2 );
SMC_CODES faSMCready2exit( void *dummy1, void *dummy2 );

// C prototypes
#ifdef sun
extern "C" long gethostid();
extern "C" int  gethostname( char *hostname, int length );
#endif

//
// program code
//
extern "C" int
fainit_()
{
  // 
  // init signal catchers
  //
  if( FARM ){
    signal( SIGINT,  faCatchSIG ); //  2
    signal( SIGQUIT, faCatchSIG ); //  3
    signal( SIGTERM, faCatchSIG ); // 15
    signal( SIGILL,  faCatchSIG ); //  4
    signal( SIGBUS,  faCatchSIG ); // 10
    signal( SIGSEGV, faCatchSIG ); // 11
  }

  //
  // node ID
  //
  unsigned int nodeID = gethostid();
#ifdef linux
  nodeID = ( nodeID & 0x0000ff00 ) >> 8;
#else
  nodeID = ( nodeID & 0x000000ff );
#endif

  //
  // host name
  //
  char *hostName = new char[30];
  gethostname( hostName, 29 );

  //
  // current date/time
  //
  time_t t = time( NULL ); 
  char date[25];
  strcpy( date, ctime( &t ) );
  date[24] = '\0';

  //
  // process id
  //
  int pid = getpid();

  // component flags
  char *rcompmask = getenv( RC_Comp_Mask_ENV );
  int compMask = 0;
  if( rcompmask ) compMask = atoi( rcompmask );
#ifdef _OUT_
  cout << "INIT: rcompmask=" << rcompmask << endl;
#endif
  
  // reco flags
  char *rrecomask = getenv( PRCMRECOMASK );
  int recoMask = 0;
  if( rrecomask ) recoMask = atoi( rrecomask );
#ifdef _OUT_
  cout << "INIT: rrecoMask=" << rrecomask << endl;
#endif
  
  // configuration
  char *configuration = getenv( PRCMRUNCFG );
  unsigned int confMask = 0;
  if( configuration ) confMask = atoi( configuration );
#ifdef _OUT_
  cout << "INIT: configuration=" << configuration << endl;
#endif

  // settings
  int runName = RCFG_GET_NAME    ( confMask );
  int runType = RCFG_GET_TYPE    ( confMask );
  int logType = RCFG_GET_LOGGING ( confMask );
  int trigger = RCFG_GET_TRIGGERS( confMask );

  // print
#ifdef _OUT_
  if( FADEB ){
    cout << "INIT: "
	 << "comp="    << compMask << " "
	 << "reco="    << recoMask << " "
	 << "name="    << runName  << " "
	 << "type="    << runType  << " "
	 << "logging=" << logType  << " "
	 << "trigger=" << trigger  << endl;
  }
#endif

  //
  // write to log-file
  //
  if( LOGFILE ){
#ifdef _OUT_
    cout << "INIT: Write to logfile=" << LOGFILE << endl;
#endif
    FILE *log = fopen( LOGFILE, "a+" );
    if( log != NULL ){
      fprintf( log, "START on %s node:%s pid:%d comp:%d reco:%d name:%d type:%d logging:%d trigger:%d",
	       date, hostName, pid, compMask, recoMask, runName, runType, logType, trigger );
      for( int i=0; i<nArg; i++ ){
	fprintf( log, " %s", vArg[i] );
      }
      fprintf( log, "\n" );
      fclose( log );
    }
  }

  //
  // rpm
  //
  int rc = GRR_OK;
  if( RPMINIT ){

    //
    // rpm/smc/rhp *************************************************************
    //
    int count = 0;
    while( 1 ){
      count++;
      
      // exit when many trails fail
      if( count > 5 ) faCatchSIG( -1 );

      //
      // rpm *******************************************************************
      //
      
      // rpm init
      if( ( rc = rpm_init( nArg, (const char **) vArg ) ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: rpm_init() rc=" << rc << endl; 
#endif
      }
      // error
      else {
#ifdef _ERR_
	cout << "INIT: rpm_init() FAILED. rc=" << rc << endl;
#endif
	// grr
	ARTE_error( GRR_SFATAL, "INIT: rpm_init() FAILED.", rc );
	
	// rpm_halt()
	rc = rpm_halt();

	// try again after some time
	sleep( 5 );
	continue;
      }
      
      // rpm open
      if( ( rc = rpm_open( RPM_REGISTER ) ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: rpm_open( RPM_REGISTER ) rc=" << rc << endl;
#endif
      }
      // error
      else {
#ifdef _ERR_
	cout << "INIT: rpm_open() FAILED. rc=" << rc << endl;
#endif
	// grr
	ARTE_error( GRR_SFATAL, "INIT: rpm_open() FAILED.", rc );
	
	// rpm_halt()
	rc = rpm_halt();

	// try again after some time
	sleep( 5 );
	continue;
      }

      // initialize pinging
      if( ( rc = rpm_bind( RPM_STATREQ, rpi_statmsgx, 0 ) ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: rpm_bind( RPM_STATREQ ) rc=" << rc << endl;
#endif
      }
      // error
      else {
#ifdef _ERR_
	cout << "INIT: rpm_bind( RPM_STATREQ ) FAILED. rc=" << rc << endl;
#endif
	// grr
	ARTE_error( GRR_SFATAL, "INIT: rpm_bind( RPM_STATREQ ) FAILED.", rc );
	
	// rpm_halt()
	rc = rpm_halt();
	
	// try again after some time
	sleep( 5 );
	continue;
      }
      
      // grr
      ARTE_error( GRR_SINFO, "INIT: RPM initialized.", 0 );

      //
      // smc *******************************************************************
      //

      // smc init
      if( ( rc = smc_init( nArg, (const char **) vArg ) ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: smc_init() rc=" << rc << endl; 
#endif
      }
      // error
      else {
#ifdef _ERR_
	cout << "INIT: smc_init() FAILED. rc=" << rc << endl;
#endif
	// grr
	ARTE_error( GRR_SFATAL, "INIT: smc_init() FAILED.", rc );
	
	// rpm_halt()
	rc = rpm_halt();

	// try again after some time
	sleep( 5 );
	continue;
      }

      // smc bind
      if( ( rc = smc_bind() ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: smc_bind() rc=" << rc << endl; 
#endif
      }
      // error
      else {
#ifdef _ERR_
	cout << "INIT: smc_bind() FAILED. rc=" << rc << endl;
#endif
	// grr
	ARTE_error( GRR_SFATAL, "INIT: smc_bind() FAILED.", rc );
		
	// rpm_halt()
	rc = rpm_halt();
	
	// try again after some time
	sleep( 5 );
	continue;
      }
    
      // smc define state transitions
      smc_add_st_fn( smc_state_atoi( "INITIALIZED" ),
		     smc_state_atoi( "CONFIGURED" ),
		     faSMCinit2conf );
      
      smc_add_st_fn( smc_state_atoi( "READY" ),
		     smc_state_atoi( "EXIT" ),
		     faSMCready2exit );
      
      //
      // smc start up
      //
      if( ( rc = smc_startup( NULL, nArg, (const char **) vArg ) ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: smc_startup() rc=" << rc << endl; 
#endif
      }
      // error
      else {
#ifdef _ERR_
	cout << "INIT: smc_startup() FAILED. rc=" << rc << endl;
#endif
	// grr
	ARTE_error( GRR_SFATAL, "INIT: smc_startup() FAILED.", rc );
	
	// rpm_halt()
	rc = rpm_halt();
	
	// try again after some time
	sleep( 5 );
	continue;
      }

      // grr
      ARTE_error( GRR_SINFO, "INIT: SMC initialized.", 0 );

      // quit loop
      break;
    }

    //
    // rhp *********************************************************************
    //
    rc = rhp_gsinit();
    if( FADEB ) ARTE_error( GRR_SDEBUG, "INIT: rhp_gsinit().", rc );
    rc = rhp_gsbind();
    if( FADEB ) ARTE_error( GRR_SDEBUG, "INIT: rhp_gsbind().", rc );

    // grr
    ARTE_error( GRR_SINFO, "INIT: RHP initialized.", 0 );

    //
    // par *********************************************************************
    //
    if( ( rc = par_sinit() ) != GRR_OK ){
      ARTE_error( GRR_SERROR, "INIT: par_sinit().", rc );
    }
    if( ( rc = par_sbind()) != GRR_OK ){
      ARTE_error( GRR_SERROR, "INIT: par_sbind().", rc );
    }

    // book
    rc = par_book( "Host Name",    PAR_T_CHAR|PAR_T_RDONLY|PAR_T_PUBLISH, 30, hostName );
    rc = par_book( "Process ID",   PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &pid );
    rc = par_book( "Run Name",     PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &runName );
    rc = par_book( "Run Type",     PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &runType );
    rc = par_book( "Logging Type", PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &logType );
    rc = par_book( "Trigger",      PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &trigger );
    rc = par_book( "Event Count",  PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &arteflag_.numevt );
    rc = par_book( "Run Number",   PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &arteflag_.numrun );
    rc = par_book( "Exp Number",   PAR_T_INT |PAR_T_RDONLY|PAR_T_PUBLISH, 1,  &arteflag_.numexp );

    // grr
    ARTE_error( GRR_SINFO, "INIT: PAR initialized.", 0 );
  }

  //
  // farm mode *****************************************************************
  //
  if( FARM ){

    //
    // init buffer
    //
    if( ( evtBuf = new char[SEG_LEN] ) != 0 ){
#ifdef _OUT_
      cout << "INIT: event buffer size=" << SEG_LEN << endl;
#endif
    }
    else {
#ifdef _ERR_
      cout << "INIT: ERROR: event buffer" << endl;
#endif
      return ARTE_ERR;
    }
    
    //
    // online running
    //
    if( FAON ){
      
      // init emg
      if( ( rc = faInitEmg() ) != GRR_OK ){
	ARTE_error( GRR_SFATAL, "INIT: emg init FAILED.", rc );
	return ARTE_ERR;
      }
    }
  }

  //  
  // init logging (definition of data type via line command)
  //
  if( DAQLOG && RPMINIT ){
    while( 1 ){
      if( ( rc = l4aLogInit( nArg, (const char **) vArg ) ) == GRR_OK ){
#ifdef _OUT_
	cout << "INIT: l4aLogInit() rc=" << rc << endl;
#endif
	break;
      }
      // error init logger
      else {
#ifdef _ERR_
	cout << "INIT: ERROR: l4aLogInit() rc=" << rc << " Trying again ..." << endl;
#endif
	ARTE_error( GRR_SERROR, "INIT: l4aLogInit(). Sleep and retry.", rc );
	sleep( 5 );
      }
    }
  }

  // grr error message
  ARTE_error( GRR_SINFO, "INIT: faInit() DONE.", 0 );
  
  // return
  return ARTE_GOOD;
}

//
// SMC functions
//
SMC_CODES faSMCinit2conf ( void *dummy1, void *dummy2 )
{
#ifdef _OUT_
  cout << "INIT: SMC INITIALIZED -> CONFIGURED" << endl;
#endif
  static int first = 1;
  SMC_CODES rc = SMC_OK;
  if( first ){
    first = 0;
    rc = SMC_WAIT;
  }
  else {
    first = 1;
    rc = SMC_OK;
  }

  // grr
  ARTE_error( GRR_SINFO, "INIT: SMC INITIALIZED -> CONFIGURED ", rc );

  return rc;
}

//
// SMC function
//
SMC_CODES faSMCready2exit( void *dummy1, void *dummy2 )
{
#ifdef _OUT_
  cout << "INIT: SMC READY -> EXIT" << endl;
#endif
  static int first = 1;
  SMC_CODES rc = SMC_OK;
  if( first ){
    first = 0;
    rc = SMC_WAIT;
  }
  else {
    first = 1;
    rc = SMC_OK;
  }

  // grr
  ARTE_error( GRR_SINFO, "INIT: READY -> EXIT ", rc );

  // exit gracefully
  faCatchSIG( 0 );

  return rc;
}
