
// *****************************************************************************
//
// source:    faLine.C
//
// type:      source code
//
// created:    7 Jul 1997
//
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// symbol:    faLine()
//
// input:     void
// output:    int  flag = 0: farm mode off (normal Arte)
//                      = 1: farm mode on  (farm I/O)
//
// called by: atinit_()
//
// purpose:   evaluate line command arguments and set globals accordingly
//
// hints:     FARM mode can be enabled or disabled by giving either:
//            DAQ file name
//            EMG partition name
//
// changes:   
//                  line command arguments to steer rpm (AG: 25-Nov-98)
//                  line command argument for delayed reading (AG: 26-Nov-98)
//                  multiple file input (AG: 27-Nov-98)
//                  daq format logging (AG: 1-Dec-98)
//                  number of emg segmens to 8 (AG: 12 Jan 99)
//                  more polishing (AG: 15 Jan 99)
//                  print version description in help (AG: 21 Jan 99)
//                  polish rpm handling (AG: 2 Feb 99)
//                  submission date of sources (AG: 3 Feb 99)
//                  increased default buffer size (AG: 11 Feb 99)
//                  introduce argument to cycle until stop (AG: 12 Feb 99)
//                  use same buffer size as everywhere else (AG: 17 Feb 99)
//                  polishing (AG: 19 Feb 99)
//                  init grr (AG: 25 Feb 99)
//                  add flag for online display (AG: 25 Mar 99)
//                  add flag to suppress log files (AG: 26 Mar 99)            
//                  add -length:<event_buffer_length> (AG: 14 May 99)
//                  change STAGE_DISK -> STAGE_PATH (AG: 26 May 99)
//                  put includes (AG: 1 Jun 99)
// 20 Jul 1999 (AG) increase default buffer size
// 22 Jul 1999 (AG) polishing
// 05 Oct 1999 (AG) polishing
// 08 Oct 1999 (AG) increase input file number to 200
// 10 Oct 1999 (AG) enable log file for online purposes
// 13 Oct 1999 (AG) improve command line argument handling
// 21 Oct 1999 (AG) polishing
// 03 Nov 1999 (AG) rename -L4Ocycle to -L4Onoblock
// 03 NOV 1999 (AG) print signal to file with -L4Ologsig:<file>
// 24 Jan 2000 (AG) introduce -L4Odummy mode to run without input files
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

// ONLINE includes
#include "stage.h"

// Arte includes
#include "arte/ArteFlag.h"
#include "farm/fa.hh"

// globals
char *LOGFILE  = 0;          // log file
char *LOGSIG   = 0;          // log file
int  FARM      = 0;          // default FARM mode
int  FAON      = 0;          // default FARM online=off
int  RPMINIT   = 0;          // init rpm
int  NOSPAR    = 1;          // default sparsification
int  FADEB     = 0;          // default NO debug mode
int  DAQLOG    = 0;          // daq format logging
int  NOBLOCK   = 0;          // no blocking of emg
int  DISPLAY   = 0;          // display
int  FAGRR     = 0;          // grr logger
char *PARNAME  = 0;          // default emg partition name
int  SEG_NUM   = 4;          // default emg number of segments
int  SEG_LEN   = 250000;     // default event buffer size
int  WAIT      = 0;          // delay for reading event
char *DBNAME   = 0;          // default data base file name
char *IN_FILE[N_IN_FILE];    // input files (NO defaults)
char *HBSTAGE  = STAGE_PATH; // stage disk
int  OLD       = 0;          // old file mode
int  nFiles    = 0;          // number of files

int  CYCLE     = 0;          // needed later

// externals
extern int  nArg;
extern char **vArg;

//
// program code
//
int
faLine()
{
  // inits
  char *files = 0;
  char *ptr   = 0;

  // evaluate line command arguments arguments
  for( int i=1; i<nArg; i++ ){
    
    // help
    if( nArg - i ){
      if( !strcmp( vArg[i], "-help" ) ||
	  !strcmp( vArg[i], "-h" )    ||
	  !strcmp( vArg[i], "-H" )       ){
#ifdef _OUT_
	cout << "**********************************" << endl;
	cout << "* compiled: " << __DATE__ << " " << __TIME__ << " *" << endl;
	cout << "**********************************" << endl;
	cout << "usage: " << vArg[0] << " <kumac> inter/batch"              
	     << endl;
	cout << "       [-help][-h][-H]"                                      
	     << endl;
	cout << "       [-debug                        (DEFAULT=" << FADEB      << ")]"
	     << endl;
	cout << "       [-noprint]                     (DEFAULT=" << arteflag_.noprint    << ")]"
	     << endl;
	cout << "       FARM offline options:"
	     << endl;
	cout << "       [-f <file1,file2,...(max 200)> (DEFAULT=none)]" 
	     << endl;
	cout << "       [-length:<event_buffer_length> (DEFAULT=" << SEG_LEN    << ")]" 
	     << endl;
	cout << "       EMG options (FARM online mode):"
	     << endl;
	cout << "       [-EMGpart:<partition>          (DEFAULT=" << PARNAME    << ")]" 
	     << endl; 
	cout << "       [-EMGnum:<segments>            (DEFAULT=" << SEG_NUM    << ")]" 
	     << endl;
	cout << "       [-EMGseg:<segments>            (DEFAULT=" << SEG_NUM    << ")]" 
	     << endl;
	cout << "       [-EMGlen:<length>              (DEFAULT=" << SEG_LEN    << ")]" 
	     << endl;
	cout << "       RPM initialization:"
	     << endl;
	cout << "       [-RPMinit                      (DEFAULT=" << RPMINIT    << ")]"
	     << endl;
	cout << "       [-Process:<process>] name      (DEFAULT=" << "none"     << ")]"
	     << endl;
	cout << "       [-Domain:<domain>] domain      (DEFAULT=" << "/OTHERS/" << ")]"
	     << endl;
	cout << "       Further options:"
	     << endl;
	cout << "       [-L4Ostage:<stage_disk>]       (DEFAULT=" << HBSTAGE    << ")]"
	     << endl;
	cout << "       [-L4Ogrr] error logger         (DEFAULT=" << FAGRR      << ")]"
	     << endl;
	cout << "       [-L4Onoblock] do NOT block emg (DEFAULT=" << NOBLOCK    << ")]"
	     << endl;
	cout << "       [-L4Odummy] one event w/o file (DEFAULT=" << CYCLE      << ")]"
	     << endl;
	cout << "       [-L4Odisplay] online display   (DEFAULT=" << DISPLAY    << ")]"
	     << endl;
	cout << "       [-L4Owait:<seconds> read delay (DEFAULT=" << WAIT       << ")]"
	     << endl;
	cout << "       [-L4Ologfile:<logfile>         (DEFAULT=" << LOGFILE    << ")]"
	     << endl;
	cout << "       [-L4Ologsig:<logsig>           (DEFAULT=" << LOGSIG     << ")]"
	     << endl;
#endif
	// terminate
	exit( ARTE_GOOD );
      }
    }

    if( nArg - i ){ 
      
      // debugging mode
      if( !strcmp( vArg[i], "-debug" ) ){
	FADEB = 1;
      }

      // debugging mode
      if( !strcmp( vArg[i], "-noprint" ) ){
	arteflag_.noprint = 1;
      }
      
      // rpm init
      if( !strcmp( vArg[i], "-RPMinit" ) ){
	RPMINIT = 1;
      }
      
      // rpm register
      if( !strncmp( vArg[i], "-Process:", 9 ) || 
          !strncmp( vArg[i], "-Domain:", 8 )     ){
	RPMINIT = 1;
      }

      // error logging
      if( !strcmp( vArg[i], "-L4Ogrr" ) ){
	FAGRR = 1;
      }
      
      // daq format logging
      if( !strcmp(  vArg[i], "-L4Atest" )           ||
          !strcmp(  vArg[i], "-L4Areal" )           ||
          !strncmp( vArg[i], "-L4Adetector:", 13 )     ){
	RPMINIT = 1;
	DAQLOG  = 1;
      }
      
      // delayed reading 
      if( !strncmp( vArg[i], "-L4Owait:", 9 ) ){
	WAIT = atoi( vArg[i] + 9 );
      }
      
      // cycling until file cycle.stop appears
      if( !strcmp( vArg[i], "-L4Onoblock" ) ){
	NOBLOCK = 1;
      }

      // cycling without blocking emg
      if( !strcmp( vArg[i], "-L4Odisplay" ) ){
	DISPLAY = 1;
      }

      // cycling without blocking emg
      if( !strcmp( vArg[i], "-L4Odummy" ) ){
	FARM  = 1;
	CYCLE = 1;
      }

      // event buffer size
      if( !strncmp( vArg[i], "-length:", 8 ) ){
	SEG_LEN = atoi( vArg[i] + 8 );
      }
      
      // emg partition name
      if( !strncmp( vArg[i], "-EMGpart:", 9 ) ){
	FARM    = 1; // farm mode
	FAON    = 1; // farm online mode
	int len = strlen( vArg[i] ) - 9 + 1;
	//cout << "-EMGpart:len=" << len << endl;
	PARNAME = new char[len];
	memset( PARNAME, 0, len );
	strcpy( PARNAME, vArg[i] + 9 );
      }
      
      // emg partition number
      if( !strncmp( vArg[i], "-EMGnum:", 8 ) ){
	SEG_NUM = atoi( vArg[i] + 8 );
      }

      // emg partition number
      if( !strncmp( vArg[i], "-EMGseg:", 8 ) ){
	SEG_NUM = atoi( vArg[i] + 8 );
      }
      
      // emg partition size
      if( !strncmp( vArg[i], "-EMGlen:", 8 ) ){
	SEG_LEN = atoi( vArg[i] + 8 );
      }

      // stage disk
      if( !strncmp( vArg[i], "-L4Ostage:", 10 ) ){
	int len = strlen( vArg[i] ) - 10 + 1;
	//cout << "-L4Ostage:len=" << len << endl;
	HBSTAGE = new char[len];
	memset( HBSTAGE,0, len );
	strcpy( HBSTAGE, vArg[i] + 10 );
      }

      // log file
      if( !strncmp( vArg[i], "-L4Ologfile:", 12 ) ){
	int len = strlen( vArg[i] ) - 12 + 1;
	//cout << "-L4Ologfile:len=" << len << endl;
	LOGFILE = new char[len];
	memset( LOGFILE, 0, len );
	strcpy( LOGFILE, vArg[i] + 12 );
      }

      // log file for sig
      if( !strncmp( vArg[i], "-L4Ologsig:", 11 ) ){
	int len = strlen( vArg[i] ) - 11 + 1;
	//cout << "-L4Ologsig:len=" << len << endl;
	LOGSIG = new char[len];
	memset( LOGSIG, 0, len );
	strcpy( LOGSIG, vArg[i] + 11 );
      }
    }
    
    // input file
    if( nArg - i - 1 ){ 
      if( !strcmp( vArg[i], "-x" ) || 
	  !strcmp( vArg[i], "-f" )    ){
	FARM = 1; // farm mode
        files = new char[strlen( vArg[i+1] ) + 1];
	//cout << strlen( vArg[i+1] ) + 1 << endl;
        strcpy( files, vArg[i+1] );
        ptr = strtok( files, "," ); // find comma
        while( ptr ){
	  //cout << strlen( ptr ) + 1 << endl;
	  //cout << ptr << endl;
          IN_FILE[nFiles] = new char[strlen( ptr ) + 1];
          strcpy( IN_FILE[nFiles], ptr );
	  //cout << "n="    << nFiles + 1 << " file=" << ptr        << endl;
          nFiles++;
          ptr = strtok( NULL, "," );
	}
        delete[] files;
      }

      // old format
      if( !strcmp( vArg[i], "-x" ) ){
	OLD = 1; // old format (no logger)
      }
    }
  }
    
  // farm mode?
  if( !FARM ) return FARM;
  
  // print globals
  cout << endl;
  cout << "SETTINGS: "
       << "debug="        << FADEB   << " "
       << "rpm_init="     << RPMINIT << " "
       << "logging="      << DAQLOG  << " "
       << "delay="        << WAIT    << " "
       << "noblock="      << NOBLOCK << " "
       << "grr="          << FAGRR   << " "
       << "logfile="      << LOGFILE << " "
       << "logsig="       << LOGSIG  << " ";

  // farm online
  if( FAON ){
    cout << "partition="    << PARNAME  << " "
	 << "number="       << SEG_NUM  << " "
	 << "length="       << SEG_LEN  << endl;
  }
  // farm offline
  else {
    cout << "old="   << OLD     << " ";
    cout << "stage=" << HBSTAGE << " ";
    cout << endl;
    for( int i=0; i<nFiles; i++ ){
      cout << "          file" << i + 1 << "=" << IN_FILE[i] << endl;;
    }
    cout << endl;
  }

  // return  
  return FARM;
}
