// 
//  Author:  Rainer Mankel                     Created:  20-Apr-99
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

// rghlcr_
// For a given domain layer, particle arrow and tracking mode, give 
// array of HITB hits suitable for track.
// The HITB row numbers are stored into a fixed-length array defined
// in the calling routine. The number of hits in array is reset to zero
// before filling. 

#include <math.h>
#include "rang/Geom.hh"
#include "rang/DomObj.hh"
#include "rang/DomLay.hh"
#include "rang/HL.hh"
#include "rang/zfpar.hh"
#include "rang/zcut.hh"
#include "rang/rgxseg.hh"
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"

extern rangerDomObj* rangerDO;   // global DO pointer

extern "C" {
  void rghlcr_(rangerDomLay* const & dl,    // domain layer
	       const int& tMode,            // tracking mode, can be
                                            //      1 - 2D x
                                            //      2 - 2D y
                                            //      3 - 3D straight
                                            //      4 - 3D mag
	       const double& z,             // z of prediction
	       const double& x,             // x
	       const double& y,             // y
	       const double& tx,            // tan theta-x
	       const double& ty,            // tan theta-y
	       const int& maxHits,          // max space in array
	       ArtePointer<HITB>* hitVec,   // pointer to hit array for output
	       int& nHits) {                // number of hits in array

    static ArtePointer<GEDE> iGede = 0;   // for caching of prediction
    static double xp = 0;
    static double yp = 0;
    static double u = 0;
    static double du = 3;

    if (!rangerDO) {
      cout << "rghlcr: rangerDO=" << rangerDO << endl;
      return;
    }

    nHits = 0;

    // loop over hits in domain layer
    for (vector<ArtePointer<HITB> >::const_iterator hi=dl->hl()->h().begin();
	 hi != dl->hl()->h().end(); ++hi) {

      if ((*hi)->gede != iGede) {  // update prediction
	switch (tMode) {
	case 1:
	  xp = x + tx * ((*hi)->gede->zmin - z);
	  u = xp;
	  du = 3;
	  break;
	case 2:
	  {
	    //rangerDomain* dlpd = dl->dom();
	    xp = rgxseg_((*hi)->gede->zmin,dl->dom());
	    //xp = rgxseg_((*hi)->gede->zmin,dlpd);
	  }
	  if (xp < -1000 || xp > 1000) {
	    cout << " () rghlcr: invalid xp=" << xp
		 << " for gede=" << (*hi)->gede
		 << " z=" << (*hi)->gede->zmin << endl;
	    continue;
	  }
	  yp = y + ty * ((*hi)->gede->zmin - z);
	  u = rangerGeom::vol( (*hi)->gede )->Cos() * xp
	    - rangerGeom::vol( (*hi)->gede )->Sin() * yp;
	  du = 3;
	  break;
	case 3:
	case 4:
	  xp = x + tx * ((*hi)->gede->zmin - z);
	  yp = y + ty * ((*hi)->gede->zmin - z);
	  iGede = (*hi)->gede;
	  du = (tMode == 3)? 3 : 10;

	  u = rangerGeom::vol( (*hi)->gede )->Cos() * xp
	    - rangerGeom::vol( (*hi)->gede )->Sin() * yp;
	  break;
	}
	iGede = (*hi)->gede;
      }	 
	
      // remove hit if already used
      if (zcut_.lrem) {
	if (zfpar_.segstor) {
	  if (((*hi)->rseg).begin() != ((*hi)->rseg).end()) continue;
	}
	else {
	  if (((*hi)->rtra).begin() != ((*hi)->rtra).end()) continue;
	}
      }

      // check the hit
      if (fabs((*hi)->x - u) > du)  continue;

      // put into array
      if (nHits < maxHits) {
	hitVec[nHits] = (*hi);
	++nHits;
      }
      else {
	arteMsg << "rghlcr: overflowing nhl, some hits skipped" << endl;
	arteMsg.warn();
	break;
      }
    }
  }
}
    



