#include "rich/dielectric.h"
#include "rich/Losses.h"


// ===========================================================================
//
//
int DiElectric::trace(Photon* p,double step) {

//
// reflexion on two surfaces
//
  if (ReflexionLoss()>1E-10) {
    if (rnd_normal()>pow(1-ReflexionLoss(),2)) return DIM_LOSS_REFLEXIONS;
    }
//
// rayleigh scattering
//
  if (rayleigh_coeff>0) {
    if (rnd_normal()>
      exp(-rayleigh_coeff*pow(p->omega(),4)*pow(epsilon(p->lambda())-1,2)*step))
      return DIM_LOSS_RAYLEIGH;
    }
//
// transmission: step in [mm] !!
//
  if (rnd_normal()>exp(-Absorption(p->lambda())*step)) {
    return DIM_LOSS_ABSORPTION;   
    }
  p->x()+=step*p->n();  
  return NO_LOSS;
  }
  
int DiElectric::Fresnel(Photon* pp,Vec3 ns,double n1,double n2) {
  Photon& p=*pp;
  
  double cosa=(p.n()*ns);
  if (cosa<0) {
    ns*=-double(1);
    cosa=-cosa;
    }
  double sina=sqrt(double(1)-cosa*cosa);
  double sinap=n1/n2*sina;

  if (sinap>=1) {    // total reflexion
    p.n()-=double(2)*cosa*ns;  // reflexion: short hand
    p.x()+=1E-10*p.n();        // move a little bit away from the surface
    return 1;
    }
  else {  
       
    double a=sqrt(n2*n2-n1*n1*sina*sina);
    double rt=(n1*cosa-a)/(n1*cosa+a);
    double rp=(n2*n2*cosa-n1*a)/(n2*n2*cosa+n1*a);
    
// ne0t= component of E perpendicular to incident plane
    Vec3   ninc_plane= cross(p.n(),ns);
    Vec3   e0=p.polarisation();
    double ne0t      =(e0*ninc_plane); 
    Vec3   e0t = ne0t*ninc_plane; 
    Vec3   e0p = e0t-e0;
    
    double b=ne0t/norm(e0);
    double r=b*b*rt*rt+(1-b*b)*rp*rp;
    
    if (rnd_normal()<=r) { // reflexion
      p.polarisation()=rt*e0t+rp*e0p;  // fresnels law
      p.n()-=double(2)*cosa*ns;        // short hand
      p.x()+=1E-10*p.n();              // move a little bit away from the surface
      return 2;
      }
    else { // transmission
      double tt=2*n1*cosa/(n1*cosa+a);
      double tp=2*n1*n2*cosa/(n2*n2*cosa-n1*a);
      p.polarisation()=tt*e0t+tp*e0p;  // fresnels law
      
      Vec3 nps=p.n()-cosa*ns;
      if (sina!=0) nps*=(sinap/sina); 
      p.n()=nps+sqrt(1-sinap*sinap)*ns; 
      
      p.x()+=1E-10*p.n(); // move a little bit away from the surface
      return 3;
      }
    }
  }
 
int DiElectric::entering(Photon* p,Vec3 ns,double nold) {
  return (Fresnel(p,ns,nold,RefractiveIndex(p->lambda()))==3);
  }
int DiElectric::leaving(Photon* p,Vec3 ns,double nold) {
  return (Fresnel(p,ns,RefractiveIndex(p->lambda()),nold)==3);
  }
  

// ----------------------------------------------------------------------

void DiElectric::init() {

//
//  constant for Raleigh-Scattering 
//
  if (ParticleDensity()<0) rayleigh_coeff=0;
  else rayleigh_coeff=1./(6*M_PI*ParticleDensity()*P_C*P_C*P_C*P_C);
  ichflag=0;
  }

void DiElectric::Cherenkov(const Track& t,const double sstep) {
  if (!ichflag) {
    ichflag=1;
  
//
//  determine constants for Cherenkov Radiation
//
    int i;
    for (i=0;i<refr.numpoints();i++) {
      double e=refr.y(i);
      if (e>epsmax) epsmax=e;
      }
    epsmax*=epsmax;  // n=sqrt(eps)
  
    double nbin=100;
    double betamin=1./sqrt(epsmax);
    double betamax=1;
    double dbeta=(betamax-betamin)/((double)(nbin-1));

    emin=lambda2energy(refr.x(refr.numpoints()-1));
    emax=lambda2energy(refr.x(0));
    int numrefr=10*(refr.numpoints()-1);
    double de=(emax-emin)/((double)numrefr);
    double e;
    int k;
  
    for (k=0;k<nbin;k++) {
      double beta=betamin+k*dbeta;
      double beta2=beta*beta;
      double sum=0;
      for (i=0;i<numrefr;i++) {
        e=emin+(i+0.5)*de;
        double f=1E9*370E2*(1.-1./(beta2*epsilon(energy2lambda(e))));
        if (f<0)f=0;
        sum+=f*de;
        }
      dncdx.append(beta,sum);
      }
    }
//
// init coordinates: z-axis parallel to particle momentum
//                   x-axis arbitrary in plane vertical to momentum
//
  track.setrotation(t.n());// x,y arbitrary, but perpendicular to \vec{p}.
  track.x0()=t.x();
//
//  calculate number of radiated photons
//
  double beta=t.p()/t.e();
  beta2=beta*beta;

  step=sstep; 
  numphot=rnd_poisson(dncdx.LinearInterpolation(beta)*step/1000.);
/*
  int k;
  for (k=0;k<dncdx.numpoint();k++) 
    printf("%lf %lf\n",dncdx.x(k),dncdx.y(k));getchar();
  double nnn=dncdx.LinearInterpolation(beta);

printf("%d %f
%f\n",numphot,beta,dncdx.LinearInterpolation(beta)*step/1000.);getchar();
*/
  nphot=0;
  }


// ----------------------------------------------------------------------
//
//  initialize geant for cherenkov simulation
//

void DiElectric::gsckov(int n,Table& efficiency) {
  double lmin=Absorption().x(0);
  double lmax=Absorption().x(Absorption().numpoints()-1);
  if (lmin>RefractiveIndex().x(0)) lmin=RefractiveIndex().x(0);
  double lm=RefractiveIndex().x( RefractiveIndex().numpoints()-1 );
  if (lm>lmax) lmax=lm;
  
  double dl=(lmax-lmin)/float(n-1);
  float *e,*a,*rindex,*eff;
  e     =new float[n];
  a     =new float[n];
  rindex=new float[n];
  eff   =new float[n];
      
  int i;
  for (i=0;i<n;i++) {
    double l=lmax-i*dl;
    e[i]     =m2GeV(l);
    double aa=Absorption().f(l);
    if (aa==0) a[i]=1E10; 
    else a[i]     =-1./(10*log(aa));
    rindex[i]=RefractiveIndex().f(l);
    eff[i]   =efficiency.f(l);
    }

  char mat[20];
  int len=(strlen(name())>20) ? 20 : strlen(name());
  memset(mat,' ',20);
  memcpy(mat,name(),len);
#ifndef NO_ARTE
  rgsdim(mat,n,e,a,eff,rindex);
#endif

  delete eff;
  delete rindex;
  delete a;
  delete e;
  }


// ----------------------------------------------------------------------

int DiElectric::radiate(Photon* p) {

  if (++nphot>=numphot) return 0;
  double e,ct2;
  double fmax=1-1./(epsmax*beta2);
  do {
    e=(emax-emin)*rnd_normal()+emin;
    ct2=1./(epsilon(energy2lambda(e))*beta2);
    } while (fmax*rnd_normal()>(1-ct2));
    
  double phi=2*M_PI*rnd_normal();
  
  p->n()     = track.RotateToOld(FromPolar(1,sqrt(ct2),phi));
  p->x()     = track.x0()+track.ez()*step*rnd_normal();
  p->e()     = e;
  
  return 1;
  }

// ----------------------------------------------------------------------

DiElectric::DiElectric(ArtePointer<CDIM> m) {
  mat=m;
  
  int i;
  for (i=sizeof(mat->name)-1;i>=0;i--) if (mat->name[i]!=' ') break;
  nname=(char*) malloc(i+2);
  memcpy(nname,mat->name,i+1);
  nname[i+1]='\0';
  
 
  ReflexionLoss()=mat->rlos;
  ParticleDensity()=mat->dens;
  absorp=Table(mat->iabs,mat->nabs);
  refr  =Table(mat->iref,mat->nref);

  init();
  }

// ----------------------------------------------------------------------

DiElectric::DiElectric(ScriptFile& f) {
  char n[256],tag[256];
  CDIM m;
  mat=ArteTable<CDIM>::push_back(m);

  ParticleDensity()=-1;
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    int err=0;
    if      (!strcmp(tag,"dielectric")) { 
      continue;
      }  
    else if      (!strcmp(tag,"name")) {   
      err=(sscanf(s,"%s",tag)!=1);
      nname=strdup(tag);
      }
    else if (!strcmp(tag,"ParticleDensity")) { 
      err=(sscanf(s,"%e",&ParticleDensity())!=1);
      }
    else if (!strcmp(tag,"ReflexionLoss")) { 
      err=(sscanf(s,"%e",&ReflexionLoss())!=1);
      }
    else if (!strcmp(tag,"absorption_length")) absorp=Table(f);
    else if (!strcmp(tag,"refractive_index")) refr=Table(f);
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "DiElectric(ScriptFile):  Unknown tag: %s %d\n",tag,err);
      Exit(0);
      }
    }    

  memset(&mat->name[0],' ',sizeof(mat->name));
  int len=(strlen(nname)>sizeof(mat->name)) ? sizeof(mat->name) :strlen(nname);
  memcpy(&mat->name[0],nname,len);

  mat->iabs=absorp.firstpoint();
  mat->nabs=absorp.numpoints();
  mat->iref=refr.firstpoint();
  mat->nref=refr.numpoints();

  init();
  }

// ----------------------------------------------------------------------

void DiElectric::print(FILE *fp) const {
  fprintf(fp,"#\ndielectric\n");
  fprintf(fp,"  name            %s\n",name());
  if (ParticleDensity()>=0)
    fprintf(fp,"  ParticleDensity  %e\n",ParticleDensity());
  fprintf(fp,"  ReflexionLoss    %e\n",ReflexionLoss());
  fprintf(fp,"#\n# absorption loss / millimeter\n absorption_length\n");
  Absorption().print(fp,0);
  fprintf(fp,"#\n refractive_index\n");
  RefractiveIndex().print(fp,0);
  fprintf(fp,"  end  # dielectric\n");
  }

DiElectric::DiElectric() {
  CDIM m;
  mat=ArteTable<CDIM>::push_back(m);
  nname=0;
  ReflexionLoss()=0;
  ParticleDensity()=-1;
  rayleigh_coeff=0;
  }

DiElectric::~DiElectric() {
  if (nname) free(nname);
  }

