
//**********************************************************************/
//   RICH reconstruction (rire), version 2.00
//
// Created: 08.03.99, by Ignacio Ari\~no. 
//        
// RichTrack: main class for the RICH reconstruction from tracks.
//  *Initialisation: from a segment (RcSegPoint or RcMatchPoint),  
//          a rec. track(RTRA), or a set of parameters
//  *To find the likelihood for a given mass hypothesis call the function
//       FindLikelihood(int masshyp,RichHits *richhits);
//  * The class Richhits has to be initialised before in order to fill it 
//      with the Rich hits. (in future versions it will be automatic)
// 
// Modified 29.11.99 by Ignacio Arin\~o. (version 3.00)
//  * Main modifications: expected background photons estimated
//    fitting a 2nd grade polynomial function to the photons which can not
//    be associated with any hypothesis.
//  * Also changed: all hypothesis now evaluated at once to save time with
//    the function FindLikelihoods(Richhits *richhits)
//
//**********************************************************************/


#ifndef _RICHRECO_HH_
#define _RICHRECO_HH_

//Marple
#include "marple/marple.hh"
#include "marple/RcMatchBuilder.hh"

#include <arte/arte.hh>

#include "CLHEP/Vector/ThreeVector.h"
#include "CLHEP/Matrix/SymMatrix.h"

//#include "rise/Corrections.h" //old
#include "rich/PhiLambda.h"

#include "rire/richhits.hh"

extern float Meanrefrindex;
extern float Nphotbeta1;


class RichTrack
{
public:
  //constructors
  RichTrack();
  RichTrack(RcSegPoint);
  RichTrack(RcMatchPoint);
  RichTrack(Hep3Vector point, Hep3Vector direction, float momentum );

  RichTrack(ArtePointer<RTRA> aprtra);

  //destructor
  ~RichTrack();
  
  void Readconstantsfromfile();
  void SetShadow();

  ///returns the number of expected photons
  double NumExpPhotons(int masshyp);


  ///finds the likelihoods for all 5 mass hypothesis
  void FindLikelihoods(RichHits *richhits);


  /// returns the relative angle between the track and a photon, 
  /// taking into account optical corrections
  double RelTheta(double hitphi,double hitlambda,int up=0);

  inline double masshyp(int hyp)
  {
    switch(hyp)
      {
      case 1: return 0.000511;  //electron
      case 2: return 0.1056584; //muon
      case 3: return 0.1395699; //pion
      case 4: return 0.493677;  //kaon
      case 5: return 0.938272;  //proton
      default: return 0;
      }
  }	

  ///return the expected Cerenkov angle for the mass hypothesis
  inline double ThetaCer(int ihyp)
    {
      double beta=1/sqrt(1+pow(masshyp(ihyp)*track->pe(),2));
      if(beta<1/Meanrefrindex) 
	return 0;
      else 
	return acos(1/(beta*Meanrefrindex));
    }

  inline Hep3Vector RichEntryPoint(){
    Hep3Vector entpoint(track->x(),track->y(),track->z());
    return entpoint;
  }

  inline Hep3Vector RichDir(){
    Hep3Vector dir(track->tx(),track->ty(),1);
    return dir.unit();
  }

  /// returns the error in phi due to the track error
  inline double SigmaPhiTrack(){ return sqrt(track->cov()(3,3))/
				   (1+track->tx()*track->tx());}
  /// returns the error in lambda due to the track error
  inline double SigmaLambdaTrack(){ return sqrt(track->cov()(4,4))/
				      (1+track->ty()*track->ty());}
  /// returns track coordinate phi
  inline double TrackPhi(){ return atan(track->tx());}

  /// returns track coordinate lambda
  inline double TrackLambda(){ return atan(track->ty());}

  /// returns the track in the RICH
  inline RcSegPoint* Track(){ return track;}

  /// returns the track momentum
  inline double Mom() { return fabs(1/track->pe());}

  //now is the same as RichDir, but will be the initial direction, 
  //when magnet is on
  inline Hep3Vector Dir(){
    Hep3Vector dir(track->tx(),track->ty(),1);
    return dir.unit();
  }

  //variables returned
  inline float Minazi(){return minazi;}
  inline float Maxazi(){return maxazi;}

  inline float ZRichMin(){return zRichmin;}
  inline float ZRichMax(){return zRichmin;}
  inline float ZMidRich(){return (zRichmax+zRichmin)/2;}
  inline float Radlength(){return zRichmax-zRichmin;}

  inline float Sigma_sph(){return Sigma_spherical;}
  inline float Sigma_crom(){return Sigma_cromatic;}
  inline float Sigma_cell(){return Sigma_cellsize;}
 

  inline double Likelihood(int i) {return lnlikelihoods[i];}

  double BackSlope(float sigmamean,float binsize,int numback,int *histoback);
  inline double BackSlope(){return backslope;}

  double back_chi(float slope, float alpha, float beta, int nbins,
		  float binsize,int *histoback);

  inline float BackAlpha(){return back_alpha;}
  inline float BackBeta(){return back_beta;}
  
  float geoeff(float theta);
  
  inline float RichTrack::geomefi(float theta)
    {
      int i=int(15*theta/maxtheta);
      return geomefis[i];
    }
  void Fillgeomefi();

private:
  RcSegPoint *track;
  double lnlikelihoods[6];

  //beam pipe shadow
  float minazi;
  float maxazi;

  //parameters to be read from ARTE tables
  float zRichmin;
  float zRichmax;

  //parameters to be read from a file
  //  double Meanrefrindex;  
  //  double Nphotbeta1;
  double Sigma_spherical;
  double Sigma_cromatic;
  double Sigma_cellsize;

  // corrections parameters
  //Corrections corr; //old


  double backslope;
  float back_alpha;
  float back_beta;

  float geomefis[15];

  float maxtheta;
  float numsigmas;


  //private function called from FindLikelihoods(...)
  double like(int ihyp, int imin, int imax, 
	      float sigmamean,double *likeback, vector<Pcand*>VCand);
  
  
};

#endif







