#include "L3_Geometry.hh"
#include "L3_Parameter.hh"

#ifdef L3_EGCS
#include <cmath>
#include <cstdio>
#include <algorithm>
#endif
#ifdef L3_GCC
#include <math.h>
#include <stdio.h>
#include <algo.h>
#endif

#ifdef L3_SLP_FRAME
#ifdef L3_EGCS
#include <cstdlib>
#endif
#ifdef L3_GCC
#include <stdlib.h>
#endif
extern "C" {
#include "l2geo.h"
}
#endif

vector<L3_Target> L3_Geometry::targetList; //list with targets (i.e. target wires)
vector<L3_Target*> L3_Geometry::targetListZ;  
float L3_Geometry::the_maxztarget;
float L3_Geometry::the_minztarget;
float L3_Geometry::the_maxxtarget;
float L3_Geometry::the_minxtarget;
float L3_Geometry::the_maxytarget;
float L3_Geometry::the_minytarget;
float L3_Geometry::the_TargetBox_xmin;
float L3_Geometry::the_TargetBox_xmax;
float L3_Geometry::the_TargetBox_ymin;
float L3_Geometry::the_TargetBox_ymax;
float L3_Geometry::the_TargetBox_zmin;
float L3_Geometry::the_TargetBox_zmax;
L3_Plane L3_Geometry::SiVD_PlaneList[7][4][4];
float L3_Geometry::SuperLayer_z[7][4];
L3_Window L3_Geometry::secs[7][4]; 
Hep3Vector L3_Geometry::the_mean_target;
int L3_Geometry::LUT[7][4][4][3];

void L3_Geometry::initialize() {
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Geometry::initialize() has been called" << endl;
#endif
  fillTargetGeometry();
  fillVDSGeometry();
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Geometry::initialize() is done" << endl;
#endif
}

void L3_Geometry::updateTarget(){
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()){
    for(vector<L3_Target>::iterator i=targetList.begin();i!=targetList.end();i++){
      printf("Target %i (%s) %f, %f, %f\n",i-targetList.begin()+1,i->name(),i->x(),i->y(),i->z());
    }
  }
#endif
#ifdef L3_EGCS
  targetList.clear();
  targetListZ.clear();
#endif
#ifdef L3_GCC
  targetList.erase(targetList.begin(),targetList.end());
  targetListZ.erase(targetListZ.begin(),targetListZ.end());
#endif
  fillTargetGeometry();
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()){
    for(vector<L3_Target>::iterator i=targetList.begin();i!=targetList.end();i++){
      printf("Target %i (%s) %f, %f, %f\n",i-targetList.begin()+1,i->name(),i->x(),i->y(),i->z());
    }
  }
#endif
}

void L3_Geometry::update() {
//in SLP it might happen that we are called again because new target positions
//were read into the SLP, so we clear all existing target wires
#ifdef L3_EGCS
  targetList.clear();
  targetListZ.clear();
#endif
#ifdef L3_GCC
  targetList.erase(targetList.begin(),targetList.end());
  targetListZ.erase(targetListZ.begin(),targetListZ.end());
#endif
  fillTargetGeometry();
  fillVDSGeometry();
}

void L3_Geometry::fillTargetGeometry() {
  readTargets();
  calculate_TargetPosition();
  calculate_TargetBox();
}

void L3_Geometry::fillVDSGeometry() {
  readPlanes();
  calculate_mean_z();
  calculate_sectors();
}

void L3_Geometry::readTargets() {
#ifdef L3_ARTE_FRAME
  for(ArteTable<GTAR>::iterator t=ArteTable<GTAR>::begin();
      t!=ArteTable<GTAR>::end();t++) {
    L3_Target wire(*t);
    if(!ArteTable<MCIN>::booked()){
      //for real data we check whether the wire is on the beam - 
      //according to D. Samtleben
      if( pow(wire.x(),2)+pow(wire.y(),2) >= 1.0 ){
#ifdef L3_VERBOSE
	if(L3_Parameter::Info()) cout << "Wire " << wire.name() << " is not on the beam...." << endl;
#endif
	continue;
      }
    }
    targetList.push_back( wire );
  }
#endif
#ifdef L3_SLP_FRAME
  for(const L2TARG *wire = l2targ_IterateInit(); wire!=NULL; wire = l2targ_IterateNext() ) {
    targetList.push_back( L3_Target(*wire) );
  }
#endif
#ifdef L3_VERBOSE
  if( L3_Parameter::Trace() ){
    cout << "L3_Geometry::readTargets: Read " << nTarget() << " target wires" << endl;
    for(vector<L3_Target>::const_iterator i = targetList.begin();i!=targetList.end();i++) {
      cout << *i;
    }
  }
#endif
}

void L3_Geometry::showGeometry()
{
  char buf[] = "!!";
  for(int i=0;i<9*7+7;i++) cout << "-"; cout << endl;
  cout << "Sector ";
  for(int sl=0;sl<7;sl++) printf("  SL%02i   ",sl+1);     
  cout << endl;
  for(int i=0;i<9*7+7;i++) cout << "-"; cout << endl;
  for(int sec=0;sec<4;sec++)
    {
      cout << "   " << sec+1 << "   ";
      for(int sl=0;sl<7;sl++){
	for(int pl=0;pl<4;pl++){
	  const L3_Plane& p = L3_Geometry::getPlane(sl,sec,pl);
	  if( p.l3nsl() == -1 ){
	    cout << "  ";
	  } else {
	    buf[0] = (atan(p.sin()/p.cos())>0) ? '+' : '-';
	    buf[1] = (fabs(p.cos())>0.5) ? 'y' : 'x';
	    cout << buf;
	  }
	}
	cout << ' ';
      }
      cout << endl;
    }
  for(int i=0;i<9*7+7;i++) cout << "-"; cout << endl;  
}

int L3_Geometry::readPlanes() {
  int foundSomething = 0;
  for (int SL=0; SL<7; SL++) {
    for (int Sec=0; Sec<4; Sec++) {
      for (int Pl=0; Pl<4; Pl++) {
	L3_Plane l3plane( 0.5, -1, (SL*16+Sec*4+Pl+1) );
	SiVD_PlaneList[SL][Sec][Pl] = l3plane;
      }
    }
  }
#ifdef L3_ARTE_FRAME
  for (ArteTable<GEDE>::iterator PIt=ArteTable<GEDE>::begin(); 
       PIt!=ArteTable<GEDE>::end(); ++PIt ) {
    ArtePointer<GEDE> Plane(*PIt);
    // Unpack SiVD GEDE-Planes, if packed
    if(Plane->sec > 65000) {
      Plane->sec = Plane->sec & 0xFFFF;
    }
    if(Plane->lay > 65000) {
      Plane->lay = Plane->lay & 0xFFFF;
    }  
    ArtePointer<GESL> agesl(Plane->gesl);
    int SL = agesl.index();
    if ( (Plane->cmp == 1) && (Plane->zfw[0] < 180) && (SL > 0 )) {
      SL = agesl->nsl;
      int l3gede = Plane.index();
      L3_Plane l3plane(Plane, SL, l3gede);
      if( SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1].l3nsl() != -1 )
	{
	  cout << "L3_Geometry::readPlanes: WARNING (sl,sec,lay)=(";
	  cout << l3plane.l3nsl() << ',' << l3plane.sec() << ',' << l3plane.lay() << ") is overwritten!!" << endl;
	}
      SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1] = l3plane;
      //cout << l3plane.l3nsl() << ' ' << l3plane.sec() << ' ' << l3plane.lay() << ' ' << l3plane.cos() << ' ' << l3plane.sin();
      //cout << ' ' << atan( l3plane.sin()/l3plane.cos() )*180.0/M_PI << ' ';
      //cout << l3plane.sxmn() << ".." << l3plane.sxmx() << "  " << l3plane.symn() << ".." << l3plane.symx() << endl;
      foundSomething = 1;
    }
  }
#endif
#ifdef L3_SLP_FRAME
  char nameSL[6];
  for(int iSL = 1; iSL < 8; iSL++) { // for each superlayer in the vertex detector
    sprintf(nameSL,"SI%02i",iSL);
    for (int Sector = 0; Sector < 4; Sector++) { // for each Sector
      int nLayers = 0;
      // get pointers to Layers
      PLANE** plArray = Geo_GetPlanes(nameSL, Sector+1, &nLayers);      
      cout << "L3: " << nameSL << ' ' << Sector+1 << ' ' << nLayers << ' ' << (void*)plArray << endl;
      if ( plArray != NULL) {
	if (nLayers != 0) {
	  foundSomething = 1;
	  for (int Layer = 0;Layer<nLayers;Layer++) { // for each Layer
	    if( Layer>=4 ){
	      cout << "Error when reading geometry, layer = " << Layer << endl;
	      continue;
	    }
            PLANE *siPl = plArray[Layer];
	    int l3gede = 16*(iSL-1)+ 4*Sector + Layer + 1;
	    L3_Plane l3plane(siPl, iSL, l3gede);
	    if( SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1].l3nsl() != -1 )
	      {
		cout << "L3_Geometry::readPlanes: WARNING (sl,sec,lay)=(";
		cout << l3plane.l3nsl() << ',' << l3plane.sec() << ',' << l3plane.lay() << ") is overwritten!!" << endl;
	      }
	    SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1] = l3plane;
	  }	    
	}
        free(plArray);
      }
    }
  }
#endif  
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) showGeometry();
#endif
  calculateLUT();
  return foundSomething;
}

void L3_Geometry::calculate_TargetPosition() {
  //calculate target position for constraints  
  if( nTarget() > 0 ) {
    the_minztarget = targetList[0].z();
    the_maxztarget = the_minztarget;
    the_minytarget = targetList[0].y();
    the_maxytarget = the_minytarget;
    the_minxtarget = targetList[0].x();
    the_maxxtarget = the_minxtarget;
    the_mean_target = targetList[0].get3center();
    for(vector<L3_Target>::const_iterator i = targetList.begin()+1;i!=targetList.end();i++)
      {
	if( i->z() < the_minztarget ) the_minztarget = i->z();
	if( i->z() > the_maxztarget ) the_maxztarget = i->z();
	if( i->y() < the_minytarget ) the_minytarget = i->y();
	if( i->y() > the_maxytarget ) the_maxytarget = i->y();
	if( i->x() < the_minxtarget ) the_minxtarget = i->x();
	if( i->x() > the_maxxtarget ) the_maxxtarget = i->x();
	the_mean_target += i->get3center();
      }
    the_mean_target *= 1.0/float(nTarget());
  }
  for(int i=0;i<nTarget();i++)
    {
      targetListZ.push_back( &(targetList[i]) );
    }
  for(int i=0;i<nTarget();i++)
    {
      int min = i;
      for(int k=i+1;k<nTarget();k++) if( targetListZ[k]->z() < targetListZ[min]->z() ) min=k;
      if( min!=i )
	{
	  L3_Target *tmp   = targetListZ[i];
	  targetListZ[i]   = targetListZ[min]; 
	  targetListZ[min] = tmp;
	}
    }  
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()){
    cout << "L3_Geometry::TargetPosition: zmintarget = " << the_minztarget << endl;
    cout << "L3_Geometry::TargetPosition: zmaxtarget = " << the_maxztarget << endl;
    cout << "L3_Geometry::TargetPosition: ymintarget = " << the_minytarget << endl;
    cout << "L3_Geometry::TargetPosition: ymaxtarget = " << the_maxytarget << endl;
    cout << "L3_Geometry::TargetPosition: xmintarget = " << the_minxtarget << endl;
    cout << "L3_Geometry::TargetPosition: xmaxtarget = " << the_maxxtarget << endl;
    cout << "L3_Geometry::TargetPosition: xtarget    = " << the_mean_target.x() << endl;
    cout << "L3_Geometry::TargetPosition: ytarget    = " << the_mean_target.y() << endl;
    cout << "L3_Geometry::TargetPosition: ztarget    = " << the_mean_target.z() << endl;
  }
#endif
}

void L3_Geometry::calculate_TargetBox() {
  the_TargetBox_xmin = 0.0 - fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_xmax = 0.0 + fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_ymin = 0.0 - fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_ymax = 0.0 + fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_zmin = the_minztarget - fabs(L3_Parameter::Offset_TargetBox_zmin());
  the_TargetBox_zmax = the_maxztarget + fabs(L3_Parameter::Offset_TargetBox_zmax());
#ifdef L3_VERBOSE
  if( L3_Parameter::Trace() ){
    cout << "L3_Geometry::TargetBox: xmin ... xmax : " << the_TargetBox_xmin;
    cout << " ... " << the_TargetBox_xmax << endl;
    cout << "L3_Geometry::TargetBox: ymin ... ymax : " << the_TargetBox_ymin;
    cout << " ... " << the_TargetBox_ymax << endl;
    cout << "L3_Geometry::TargetBox: zmin ... zmax : " << the_TargetBox_zmin;
    cout << " ... " << the_TargetBox_zmax << endl;
  }
#endif
}

void L3_Geometry::calculate_mean_z() {
  for (int layer=0; layer<7; layer++) {
    for (int i=0; i<4; i++) {
      const L3_Plane& Plane1 = getPlane(layer,i,0);
      const L3_Plane& Plane2 = getPlane(layer,i,1);
      const L3_Plane& Plane3 = getPlane(layer,i,2);
      const L3_Plane& Plane4 = getPlane(layer,i,3);
      float zmin = +10000.0;
      float zmax = -10000.0;
      if (Plane1.zfw(0) < zmin && Plane1.l3nsl() != -1) {
	zmin = Plane1.zfw(0);
      }
      if (Plane2.zfw(0) < zmin && Plane2.l3nsl() != -1) {
	zmin = Plane2.zfw(0);
      }
      if (Plane3.zfw(0) < zmin && Plane3.l3nsl() != -1) {
	zmin = Plane3.zfw(0);
      }
      if (Plane4.zfw(0) < zmin && Plane4.l3nsl() != -1) {
	zmin = Plane4.zfw(0);
      }
      if (Plane1.zfw(0) > zmax && Plane1.l3nsl() != -1) {
	zmax = Plane1.zfw(0);
      }
      if (Plane2.zfw(0) > zmax && Plane2.l3nsl() != -1) {
	zmax = Plane2.zfw(0);
      }
      if (Plane3.zfw(0) > zmax && Plane3.l3nsl() != -1) {
	zmax = Plane3.zfw(0);
      }
      if (Plane4.zfw(0) > zmax && Plane4.l3nsl() != -1) {
	zmax = Plane4.zfw(0);
      }
      SuperLayer_z[layer][i] = 0.5 * (zmin + zmax);
    }
  }
}

void L3_Geometry::calculate_sectors() {
  const float d = 0.1; //allow for 1 mm deviation
  for(int sl=0;sl<7;sl++)
    {
      for(int sec=0;sec<4;sec++)
	{
	  int found = 0;
	  float xmin=0,xmax=0,ymin=0,ymax=0;
	  for(int pl=0;pl<4;pl++)
	    {
	      const L3_Plane& p = getPlane(sl,sec,pl);
	      if(p.l3nsl()!=-1){
		found++;
		if( found == 1 )
		  {
		    xmin = p.sxmn();
		    xmax = p.sxmx();
		    ymin = p.symn();
		    ymax = p.symx();
		  }
		else
		  {
		    if( p.sxmn() < xmin ) xmin = p.sxmn();
		    if( p.sxmx() > xmax ) xmax = p.sxmx();
		    if( p.symn() < ymin ) ymin = p.symn();
		    if( p.symx() > ymax ) ymax = p.symx();
		  }
	      }		
	    }
	  if(found>0)
	    secs[sl][sec] = L3_Window(xmin-d,xmax+d,ymin-d,ymax+d,SLz(sl,sec));
	  else
	    secs[sl][sec] = L3_Window(-1000,-1000,-1000,-1000,-1000);
	  //#ifdef L3_VERBOSE
	  //cout << "Sector [" << sl << "][" << sec << "]:" << endl;
	  //cout << secs[sl][sec];
	  //#endif
	}
    }
}

int L3_Geometry::NPNPorNPPN(int sl,int sec){
  for(int p1=0;p1<2;p1++){
    const L3_Plane& pl1 = getPlane(sl,sec,p1);
    if( pl1.l3nsl() == -1 ) continue;
    int sign1 = (fabs(pl1.cos())>0.5) ? -1 : 1;
    for(int p2=p1+1;p2<4;p2++){
      const L3_Plane& pl2 = getPlane(sl,sec,p2);
      if( pl2.l3nsl() == -1 ) continue;
      int sign2 = (fabs(pl2.cos())>0.5) ? -1 : 1;
      if( sign1==sign2 ){
	if( (p2-p1)==2 ) //planes with the same stereo angle are separated by 2 planes
	  return +1; //NPNP
	else
	  return -1; //NPPN    
      }
    }
  } 
  return 0; //no decision
}

void L3_Geometry::calculateLUT()
{
  int Default = 0;
  for(int sl=0;Default==0 && sl<7;sl++) for(int sec=0;Default==0 && sec<4;sec++) Default = NPNPorNPPN(sl,sec);
  if( L3_Parameter::Info() ){
    switch(Default){
    case +1 : cout << "L3_Geometry: NPNP" << endl; break;
    case -1 : cout << "L3_Geometry: NPPN" << endl; break;
    default : cout << "L3_Geometry: ????" << endl; break;
    }
  }
  for(int sl=0;sl<7;sl++) for(int sec=0;sec<4;sec++){
    int type = NPNPorNPPN(sl,sec);
    if(type==0) type=Default;
    for(int pl=0;pl<2;pl++){
      if( type==1 ){  //NPNP
	LUT[sl][sec][pl][0] = 2+pl; //stereo plane
	LUT[sl][sec][pl][1] = 1-pl; //1st other plane
	LUT[sl][sec][pl][2] = 3-pl; //2nd other plane
	LUT[sl][sec][pl+2][0] = pl; //stereo plane
	LUT[sl][sec][pl+2][1] = 1-pl; //1st other plane
	LUT[sl][sec][pl+2][2] = 3-pl; //2nd other plane
      } else {        //NPPN
	LUT[sl][sec][pl][0] = 3-pl; //stereo plane
	LUT[sl][sec][pl][1] = 1-pl; //1st other plane
	LUT[sl][sec][pl][2] = 2+pl; //2nd other plane
	LUT[sl][sec][pl+2][0] = 1-pl; //stereo plane
	LUT[sl][sec][pl+2][1] = pl; //1st other plane
	LUT[sl][sec][pl+2][2] = 3-pl; //2nd other plane
      }
    }
  }    
}

const L3_Target& L3_Geometry::ClosestTarget( Hep3Vector v )
{
  const L3_Target *target = &L3_Geometry::getTarget(0);  
  float dd = (target->get3center()-v).mag();
  for(int i=1;i<L3_Geometry::nTarget();i++)
    {
      const L3_Target *tmp = &L3_Geometry::getTarget(i);
      float newdd = (tmp->get3center()-v).mag();
      if( newdd < dd )
	{
	  target = tmp;
	  dd = newdd;
	}
    }
  return *target;
}


