#ifndef __WAX_RTRELATION_HH__
#define __WAX_RTRELATION_HH__

#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifndef __STRING__
#include <string>
#endif
#ifndef __OTR_DB_H__
#include "otr_db.h"
#endif

class waxParization ;
class TH1 ;
class rtcalRTProjection ;

class waxRTRelation
{
public:
  // typedefs, embedded classes
  enum EPitch { kPitchFive=0, kPitchTen=1 } ;
  enum EParType { kTofR, kRofT } ;
  typedef float DriftRadius_t ;
  typedef float DriftTime_t ;
  typedef float XValue ;
  typedef float YValue ;
  class RTResult {
  public:
    RTResult() : fT(0),fR(0),fdTdR(0),fSigmaR(0) {}
    DriftTime_t   fT ;
    DriftRadius_t fR ;
    float         fdTdR ;
    DriftRadius_t fSigmaR ;
  } ;
  
  // constructors, destructor, assignment
  waxRTRelation() ;
  waxRTRelation(int key, const string& filename) ;
  waxRTRelation(LedaTableIterator<class OTR_r_t>&) ;
  waxRTRelation(int ledakey, EPitch pitch, EParType partype,
		int n_rt, int n_res) ;
  ~waxRTRelation() ;

  // setters and getters
  DriftRadius_t GetRMin() const { return fgRMin[fEPitch] ; }
  DriftRadius_t GetRMax() const { return fgRMax[fEPitch] ; }
  int GetLedaKey() const { return fLedaKey ; }
  unsigned char GetEPitch() const { return fEPitch ; }
  unsigned char GetEParType() const { return fEParType ; }

  // other inline methods
  inline RTResult GetRTResultFromR(const DriftRadius_t r) const ;
  inline RTResult GetRTResultFromT(const DriftTime_t t) const ;
  int GetDSRIndex() const { return fLedaKey ; }
  
  // other methods
  RTResult CalcRTResultFromR(const DriftRadius_t r) const ;
  RTResult CalcRTResultFromT(const DriftTime_t t) const ;

  // IO
  void Read(const string& filename) ;
  void Write(int format, const string& filename) const ;
  void ReadDB() ;
  void WriteDB() const ;

private:
  inline int GetBinFromR(const DriftRadius_t r) const ;
  inline int GetBinFromT(const DriftTime_t r) const ;
  static int GetNTableBins() { return fgNTableBins ; }
  void FillRTTable() ;

private: // stuff called from friend class rtcalRTProjection
  friend class otrcalRTProjection ; // the calibration is friend
  waxParization* GetTParization() { return fTParization ; }
  waxParization* GetSParization() { return fSParization ; }
  RTResult CalcRTResultFromX(const XValue x) const ;
  RTResult CalcRTResultFromY(const YValue y) const ;
  inline int GetBinFromY(const YValue y) const ;
  inline int GetBinFromX(const XValue x) const ; 
  XValue GetXMin() const { return fXMin ; }
  XValue GetXMax() const { return fXMax ; }
  void ChangeXBounds(XValue xmin, XValue xmax) ;
  inline RTResult GetRTResultFromX(const XValue x) const ;
  DriftRadius_t GetAverageSigmaR() const ;
private:
  // statics
  static const DriftRadius_t fgRMin[2] ; // nominal minimum driftradius
  static const DriftRadius_t fgRMax[2] ; // nominal maximum driftradius
  static const DriftTime_t fgTMin ; // nominal minimum drifttime
  static const DriftTime_t fgTMax ; // nominal maximum drifttime
  // others
  unsigned int fLedaKey ;       // key of this rt-relation in leda table
  unsigned int fEPitch ;        // layer pitch (5 or 10 mm)
  unsigned int fEParType ;      // type of parameterization (t(r) or r(t)) 
  waxParization* fTParization ; // parametrization of t(r) or r(t)  
  waxParization* fSParization ; //         "          V(r) or V(t) in cm^2
  XValue fXMin ;  // the minimum driftradius/time
  XValue fXMax ;  // the maximum driftradius/time
  
  // temporary data (tables for fast retrieval)
  static int fgNTableBins ; // number of bins in tables
  std::vector<RTResult> ftRTTable ;
  std::vector<int> ftMapTable ;
  YValue fYMin ;  // the minimum drifttime/radius
  YValue fYMax ;  // the maximum drifttime/radius
} ;

/*************** inlines ********************/

inline int waxRTRelation::GetBinFromY(const YValue y) const 
{
  int bin = int(fgNTableBins*(y-fYMin)/(fYMax-fYMin) + 0.5) ;
  if(bin>=fgNTableBins) bin=fgNTableBins-1 ;
  else if(bin<0) bin=0;
  return ftMapTable[bin] ;
}

inline int waxRTRelation::GetBinFromX(const XValue x) const 
{
  int bin = int(fgNTableBins*(x-fXMin)/(fXMax-fXMin) + 0.5) ;
  if(bin>=fgNTableBins) bin=fgNTableBins-1 ;
  else if(bin<0) bin=0;
  return bin ;
}

inline int waxRTRelation::GetBinFromT(const DriftTime_t t) const
{
  if(fEParType==kTofR) return GetBinFromY(t) ;
  return GetBinFromX(t) ;
}

inline int waxRTRelation::GetBinFromR(const DriftRadius_t r) const 
{
  if(fEParType==kTofR) return GetBinFromX(r) ;
  return GetBinFromY(r) ;
}

inline waxRTRelation::RTResult 
waxRTRelation::GetRTResultFromR(const DriftRadius_t r) const 
{
  RTResult rt = ftRTTable[GetBinFromR(r)] ;
  rt.fT += rt.fdTdR*(r-rt.fR) ;
  rt.fR = r ; 
  return rt;
}

inline waxRTRelation::RTResult 
waxRTRelation::GetRTResultFromT(const DriftTime_t t) const 
{
  RTResult rt = ftRTTable[GetBinFromT(t)] ;
  rt.fR += (t-rt.fT)/rt.fdTdR ;
  rt.fT = t ;
  return rt;
}

inline waxRTRelation::RTResult 
waxRTRelation::GetRTResultFromX(const XValue x) const 
{
  if(fEParType==kTofR) return GetRTResultFromR(x) ;
  return GetRTResultFromT(x) ;
}

#endif
