#include <fstream>
#include <iomanip>
#include "wax/RTRelation.hh"
#include "wax/Parization.hh"


//__________________________________________________________________
// waxRTRelation
//
// Interface class for the rt-relations
//
// Author: Wouter Hulsbergen
//__________________________________________________________________

const waxRTRelation::DriftRadius_t
waxRTRelation::fgRMin[2] = {0.00125,0.00125 } ;//cm

const waxRTRelation::DriftRadius_t
waxRTRelation::fgRMax[2] = {0.29,0.58} ;//cm

const waxRTRelation::DriftTime_t
waxRTRelation::fgTMin = 0 ;

const waxRTRelation::DriftTime_t
waxRTRelation::fgTMax = 256 ;

int waxRTRelation::fgNTableBins = 1000 ;

waxRTRelation::waxRTRelation() 
  : fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0) {}


waxRTRelation::waxRTRelation(int key, const string& filename) 
  : fLedaKey(key),fEParType(kTofR),fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0)
{
  fTParization = waxParization::Create() ;
  fSParization = waxParization::Create() ;
  Read(filename) ;
  fLedaKey = key ;
}

waxRTRelation::waxRTRelation(LedaTableIterator<class OTR_r_t>& ledaIt)
  : fLedaKey(ledaIt->irt),fEParType(kTofR),fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0)
{
  fTParization = waxParization::Create() ;
  fSParization = waxParization::Create() ;
  ReadDB() ; 
}

waxRTRelation::waxRTRelation(int ledakey, EPitch epitch, EParType partype,
			     int n_rt, int n_res) :
  fLedaKey(ledakey), fEPitch(epitch), fEParType(partype)
{
  if(partype==kTofR) {
    fTParization = waxParization::Create(GetRMin(),GetRMax(),3,n_rt,150) ;
    fSParization = waxParization::Create(GetRMin(),GetRMax(),0,n_res,0.2) ;
  } else {
    fTParization = waxParization::Create(fgTMin,fgTMax,3,n_rt,0.2) ;
    fSParization = waxParization::Create(fgTMin,fgTMax,0,n_res,0.2) ;
  }
  fXMin = fTParization->GetXMin() ;
  fXMax = fTParization->GetXMax() ;
  FillRTTable() ;
}

waxRTRelation::~waxRTRelation() 
{
  delete fTParization ;
  delete fSParization ;
} 

void waxRTRelation::ChangeXBounds(XValue xmin, XValue xmax) {
  fXMin = xmin ; fXMax = xmax ;
  fTParization->SetXBounds(xmin,xmax) ;
  fSParization->SetXBounds(xmin,xmax) ;
}

void
waxRTRelation::ReadDB()
{
  LedaPointer<class OTR_r_t> ledartIt(OTR_r_t->get->key(fLedaKey)) ;
  fEPitch = ledartIt->pitch ;
  HepVector tpar(ledartIt->n_rt) ;
  for(int i=0; i<tpar.num_row(); i++) tpar[i] = ledartIt->par_rt[i] ;
  HepVector spar(ledartIt->n_res) ;
  for(int i=0; i<spar.num_row(); i++) spar[i] = ledartIt->par_res[i] ;
//   fTParization = waxParization::Create(GetXMin(),GetXMax(),int(ledartIt->par_rt[19]),tpar) ;
//   fSParization = waxParization::Create(GetXMin(),GetXMax(),int(ledartIt->par_res[19]),spar) ;
  FillRTTable() ;
  std::cout << "waxRTRelation::ReadDB() " << GetXMax() << " " 
       << fLedaKey << std::endl ;
}

void
waxRTRelation::WriteDB() const
{
  std::cout << "waxRTRelation::WriteDB()" << std::endl ;
  LedaPointer<class OTR_r_t> ledartIt(OTR_r_t->add->key(fLedaKey)) ;
  std::cout << "LedaPointer<class OTR_r_t> created ... " << flush 
       << ledartIt->n_rt << std::endl ;
  //ledartIt->itr = fLedaKey ;
  ledartIt->pitch = fEPitch ;
  HepVector tpar = fTParization->GetPars() ;
  ledartIt->n_rt = tpar.num_row() ;
  for(int i=0; i<tpar.num_row(); i++) ledartIt->par_rt[i] = tpar[i] ;
  ledartIt->par_rt[19] = fTParization->GetDegree() ;
  HepVector spar = fSParization->GetPars() ;
  ledartIt->n_res = spar.num_row() ;
  for(int i=0; i<spar.num_row(); i++) ledartIt->par_res[i] = spar[i] ;
  ledartIt->par_res[19] = fSParization->GetDegree() ;
  std::cout << "End of waxRTRelation::WriteDB()" << std::endl ;
}

void
waxRTRelation::Read(const string& filename) 
{
  std::ifstream ifs(filename.c_str()) ;
  if(ifs) {
    int format ;
    ifs >> format ;
    //std::cout << "Reading from file:" << filename << " " << format << std::endl ;
    switch(format) { 
    case 0:
      float t0;
      ifs >> fXMin >> fXMax >> t0;
      if(fXMax>0.4) fEPitch=kPitchTen;
      else          fEPitch=kPitchFive;
      break;
    case 1:
    case 2:
      ifs >> fLedaKey >> fEPitch ;
      fXMin = fgRMin[fEPitch] ; fXMax = fgRMax[fEPitch] ;
      break;
    case 3:
      ifs >> fLedaKey >> fEPitch >> fEParType ;
      ifs >> fXMin >> fXMax ;
      break;
    case 4:
      ifs >> fLedaKey >> fEPitch >> fEParType ;
      break;
    default:
      std::cout << "waxRTRelation::Read(...): " 
	   << "file format "<< format << " not supported !" << std::endl ;  
      break;
    }
    if(format <=3) {
      fTParization->Read(fXMin,fXMax,format,ifs) ;
      fSParization->Read(fXMin,fXMax,format,ifs) ;
    } else {
      fTParization->Read(format,ifs) ;
      fSParization->Read(format,ifs) ;
      fXMin = fTParization->GetXMin() ;
      fXMax = fTParization->GetXMax() ;
    }
//     std::cout << fLedaKey << " " << fEPitch << " " << fXMin << " "<< fXMax << std::endl 
// 	 << "rtknots: " << fTParization->GetKnots()
//     	 << "rtpars: " << fTParization->GetPars() << std::endl ;
  } else {
    std::cout << "waxRTRelation::Read(...) : No such file " << filename << std::endl ;
  }
  ifs.close() ;
  FillRTTable() ;
}

void
waxRTRelation::Write(int format, const string& filename) const 
{
  //std::cout << "waxRTRelation::Write " << format << " " << filename << std::endl ;
  std::ofstream ofs(filename.c_str()) ;
  switch(format) { 
  case 0:
    ofs << fXMin << " " << fXMax << " " << 0 << std::endl ;
    break ;
  case 1:
  case 2:
    ofs << format  << " " << fLedaKey << " " << fEPitch << std::endl ;
    break;
  case 3:
    ofs << format  << " " << fLedaKey << " " << fEPitch << " " 
	<< fEParType << " " 
	<< setw(11) << setprecision(4) << GetXMin() 
	<< setw(11) << setprecision(4) << GetXMax() << std::endl ;
    break;
  case 4:
    ofs << format  << " " << fLedaKey << " " << fEPitch << " " 
	<< fEParType << std::endl ;
    break ;
  default:
    break ;
  }
  fTParization->Write(format,ofs) ;
  fSParization->Write(format,ofs) ;
}

waxRTRelation::RTResult
waxRTRelation:: CalcRTResultFromX(float x) const 
{
  RTResult rt;
  waxFunctionValue fv = fTParization->CalcFunction(x) ;
  if(fEParType==kTofR) {
    rt.fR      = fv.fX ;
    rt.fT      = fv.fY ;
    rt.fdTdR   = fv.fdYdX ;
  } else {
    rt.fR      = fv.fY ;
    rt.fT      = fv.fX ;
    rt.fdTdR   = 1/fv.fdYdX ;
  }
  rt.fSigmaR = sqrt(fSParization->CalcValue(x)) ;
  return rt ;
}

waxRTRelation::RTResult
waxRTRelation::CalcRTResultFromY(float y) const 
{
  RTResult rt;
  float precision = 0.1 ; // channel
  waxFunctionValue fv = fTParization->CalcInverse(y,precision) ;
  if(fEParType==kTofR) {
    rt.fR      = fv.fX ;
    rt.fT      = fv.fY ;
    rt.fdTdR   = fv.fdYdX ;
  } else {
    rt.fR      = fv.fY ;
    rt.fT      = fv.fX ;
    rt.fdTdR   = 1/fv.fdYdX ;
  }
  rt.fSigmaR = sqrt(fSParization->CalcValue(fv.fX)) ;
  return rt ;
}

waxRTRelation::RTResult
waxRTRelation:: CalcRTResultFromR(DriftRadius_t r) const
{
  if(fEParType==kTofR) return CalcRTResultFromX(r) ;
  return CalcRTResultFromY(r) ;
}

waxRTRelation::RTResult
waxRTRelation:: CalcRTResultFromT(DriftTime_t t) const
{
  if(fEParType==kTofR) return CalcRTResultFromY(t) ;
  return CalcRTResultFromX(t) ;
}

void waxRTRelation::FillRTTable()
{  
  //std::cout << "waxRTRelation::FillRTTable()" << std::endl ;
  ftRTTable.resize(GetNTableBins()) ;
  ftMapTable.resize(GetNTableBins()) ;

  // Fill table rtresults 
  for(int xbin=0; xbin<GetNTableBins() ; xbin++) {
    XValue x = GetXMin() + xbin*(GetXMax()-GetXMin())/GetNTableBins() ;
    ftRTTable[xbin] = CalcRTResultFromX(x) ;
//     std::cout << setw(4) << xbin 
// 	 << setw(10) << setprecision(3) << x
// 	 << setw(10) << setprecision(3) << ftRTTable[xbin].fT
// 	 << setw(10) << setprecision(3) << ftRTTable[xbin].fR 
// 	 << setw(10) << setprecision(3) << ftRTTable[xbin].fdTdR << std::endl ;
  } 
  
  // Fill map table
  if(fEParType==kTofR) {
    fYMax = ftRTTable.front().fT ;
    fYMin = ftRTTable.back().fT ;
  } else {
    fYMax = ftRTTable.front().fR ;
    fYMin = ftRTTable.back().fR ;
  }
  if(fYMax < fYMin) swap(fYMax,fYMin) ;
  float deltaY = (fYMax-fYMin)/fgNTableBins ;
  // too slow ? .... use sign deltaT !
  float precision = fabs(0.1*deltaY) ;
  for(int biny(0); biny<fgNTableBins; biny++) {
    float y = fYMin + biny*deltaY ;
    float x = fTParization->CalcInverse(y,precision).fX ;
    ftMapTable[biny] = GetBinFromX(x) ;
  }  
  //std::cout << "End of waxRTRelation::FillRTTable()" << std::endl ;
}

float waxRTRelation::GetAverageSigmaR() const
{
  float min,max ;
  min=max= ftRTTable[0].fSigmaR ;  
  float sum(0) ;
  for(int xbin=0; xbin<GetNTableBins() ; xbin++) {
    float sigma= ftRTTable[xbin].fSigmaR ;
    sum += sigma ;
    if     (min<sigma ) min = sigma ;
    else if(max>sigma ) max = sigma ;
  }
  return sum/GetNTableBins() ;
  //return 0.5*(min+max) ;
}
